# -*- Mode: Python; test-case-name: whipper.test.test_image_cue -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

"""
Read .cue files.

See http://digitalx.org/cuesheetsyntax.php
"""

import re

from whipper.common import common
from whipper.image import table

import logging
logger = logging.getLogger(__name__)

_REM_RE = re.compile(r"^REM\s(\w+)\s(.*)$")
_PERFORMER_RE = re.compile(r"^PERFORMER\s(.*)$")
_TITLE_RE = re.compile(r"^TITLE\s(.*)$")

_FILE_RE = re.compile(r"""
    ^FILE                 # FILE
    \s+"(?P<name>.*)"     # 'file name' in quotes
    \s+(?P<format>\w+)$   # format (WAVE/MP3/AIFF/...)
""", re.VERBOSE)

_TRACK_RE = re.compile(r"""
    ^\s+TRACK            # TRACK
    \s+(?P<track>\d\d)   # two-digit track number
    \s+(?P<mode>.+)$    # mode (AUDIO, MODEx/2xxx, ...)
""", re.VERBOSE)

_INDEX_RE = re.compile(r"""
    ^\s+INDEX   # INDEX
    \s+(\d\d)   # two-digit index number
    \s+(\d\d)   # minutes
    :(\d\d)     # seconds
    :(\d\d)$    # frames
""", re.VERBOSE)


class CueFile:
    """Represent a .cue file as an object."""
    logCategory = 'CueFile'

    def __init__(self, path):
        """
        Init CueFile.

        :param path: path to track
        :type path: str
        """
        assert isinstance(path, str), "%r is not str" % path

        self._path = path
        self._rems = {}
        self._messages = []
        self.leadout = None
        self.table = table.Table()

    def parse(self):
        state = 'HEADER'
        currentFile = None
        currentTrack = None
        counter = 0

        logger.info('parsing .cue file %r', self._path)
        with open(self._path) as f:
            content = f.readlines()
        for number, line in enumerate(content):
            line = line.rstrip()

            m = _REM_RE.search(line)
            if m:
                tag = m.expand('\\1')
                value = m.expand('\\2')
                if state != 'HEADER':
                    self.message(number, 'REM %s outside of header' % tag)
                else:
                    self._rems[tag] = value
                continue

            # look for FILE lines
            m = _FILE_RE.search(line)
            if m:
                counter += 1
                filePath = m.group('name')
                fileFormat = m.group('format')
                currentFile = File(filePath, fileFormat)

            # look for TRACK lines
            m = _TRACK_RE.search(line)
            if m:
                if not currentFile:
                    self.message(number, 'TRACK without preceding FILE')
                    continue

                state = 'TRACK'

                trackNumber = int(m.group('track'))

                logger.debug('found track %d', trackNumber)
                currentTrack = table.Track(trackNumber)
                self.table.tracks.append(currentTrack)
                continue

            # look for INDEX lines
            m = _INDEX_RE.search(line)
            if m:
                if not currentTrack:
                    self.message(number, 'INDEX without preceding TRACK')
                    print('ouch')
                    continue

                indexNumber = int(m.expand('\\1'))
                minutes = int(m.expand('\\2'))
                seconds = int(m.expand('\\3'))
                frames = int(m.expand('\\4'))
                frameOffset = int(frames
                                  + seconds * common.FRAMES_PER_SECOND
                                  + minutes * common.FRAMES_PER_SECOND * 60)

                logger.debug('found index %d of track %r in %r:%d',
                             indexNumber, currentTrack, currentFile.path,
                             frameOffset)
                # FIXME: what do we do about File's FORMAT ?
                currentTrack.index(indexNumber,
                                   path=currentFile.path, relative=frameOffset,
                                   counter=counter)
                continue

    def message(self, number, message):
        """
        Add a message about a given line in the cue file.

        :param message: a text line in the cue sheet
        :type message: str
        :param number: line number, counting from 0
        :type number: int
        """
        self._messages.append((number + 1, message))

    def getTrackLength(self, track):
        # returns track length in frames, or -1 if can't be determined and
        # complete file should be assumed
        # FIXME: this assumes a track can only be in one file; is this true ?
        i = self.table.tracks.index(track)
        if i == len(self.table.tracks) - 1:
            # last track, so no length known
            return -1

        thisIndex = track.indexes[1]  # FIXME: could be more
        nextIndex = self.table.tracks[i + 1].indexes[1]  # FIXME: could be 0

        c = thisIndex.counter
        if c is not None and c == nextIndex.counter:
            # they belong to the same source, so their relative delta is length
            return nextIndex.relative - thisIndex.relative

        # FIXME: more logic
        return -1

    def getRealPath(self, path):
        """
        Translate the .cue's FILE to an existing path.

        :param path: path to track
        :type path: str
        """
        return common.getRealPath(self._path, path)


class File:
    """Represent a FILE line in a cue file."""

    def __init__(self, path, file_format):
        """
        Init File.

        :param path: path to track
        :type path: str
        """
        assert isinstance(path, str), "%r is not str" % path

        self.path = path
        self.format = file_format

    def __repr__(self):
        return '<File %r of format %s>' % (self.path, self.format)



# -*- Mode: Python; test-case-name: whipper.test.test_image_table -*-
# vi:si:et:sw=4:sts=4:ts=4

from os import environ
from shutil import rmtree
from tempfile import mkdtemp
from whipper.image import table
from whipper.common import config

from whipper.test import common as tcommon


def h(i):
    return "0x%08x" % i


class TrackTestCase(tcommon.TestCase):

    def testRepr(self):
        track = table.Track(1)
        self.assertEqual(repr(track), "<Track 01>")

        track.index(1, 100)
        self.assertTrue(repr(track.indexes[1]).startswith('<Index 01 '))


class LadyhawkeTestCase(tcommon.TestCase):
    # Ladyhawke - Ladyhawke - 0602517818866
    # contains 12 audio tracks and one data track
    # CDDB has been verified against freedb:
    #   https://web.archive.org/web/20200331130804/http://www.freedb.org/freedb/misc/c60af50d
    #   https://web.archive.org/web/20200331130829/http://www.freedb.org/freedb/jazz/c60af50d
    # AccurateRip URL has been verified against EAC's, using wireshark

    def setUp(self):
        self.table = table.Table()

        for i in range(12):
            self.table.tracks.append(table.Track(i + 1, audio=True))
        self.table.tracks.append(table.Track(13, audio=False))

        offsets = [0, 15537, 31691, 50866, 66466, 81202, 99409,
                   115920, 133093, 149847, 161560, 177682, 207106]
        t = self.table.tracks
        for i, offset in enumerate(offsets):
            t[i].index(1, absolute=offset)

        self.assertFalse(self.table.hasTOC())

        self.table.leadout = 210385

        self.assertTrue(self.table.hasTOC())
        self.assertEqual(self.table.tracks[0].getPregap(), 0)

    def testCDDB(self):
        self.assertEqual(self.table.getCDDBDiscId(), "c60af50d")

    def testMusicBrainz(self):
        # output from mb-submit-disc:
        # https://musicbrainz.org/cdtoc/attach?toc=1+12+195856+150+
        # 15687+31841+51016+66616+81352+99559+116070+133243+149997+161710+
        # 177832&tracks=12&id=KnpGsLhvH.lPrNc1PBL21lb9Bg4-
        # however, not (yet) in MusicBrainz database

        # setup to test if MusicBrainz submit URL is hardcoded to use https
        env_original = environ.get('XDG_CONFIG_HOME')
        tmp_conf = mkdtemp(suffix='.config')
        # HACK: hijack env var to avoid overwriting user's whipper config file
        # This works because directory.config_path() builds the location where
        # whipper's conf will reside based on the value of env XDG_CONFIG_HOME
        environ['XDG_CONFIG_HOME'] = tmp_conf
        self.config = config.Config()
        self.config._parser.add_section('musicbrainz')
        self.config._parser.set('musicbrainz', 'server',
                                'http://musicbrainz.org')
        self.config.write()
        self.assertEqual(self.table.getMusicBrainzSubmitURL(),
                         "http://musicbrainz.org/cdtoc/attach?toc=1+12+1958"
                         "56+150+15687+31841+51016+66616+81352+99559+116070+13"
                         "3243+149997+161710+177832&tracks=12&id=KnpGsLhvH.lPr"
                         "Nc1PBL21lb9Bg4-")
        # HACK: continuation - restore original env value (if defined)
        if env_original is not None:
            environ['XDG_CONFIG_HOME'] = env_original
        else:
            environ.pop('XDG_CONFIG_HOME', None)
        self.assertEqual(self.table.getMusicBrainzDiscId(),
                         "KnpGsLhvH.lPrNc1PBL21lb9Bg4-")
        rmtree(tmp_conf)

    def testAccurateRip(self):
        self.assertEqual(self.table.accuraterip_ids(), (
            "0013bd5a", "00b8d489"))
        self.assertEqual(self.table.accuraterip_path(),
                         "a/5/d/dBAR-012-0013bd5a-00b8d489-c60af50d.bin")

    def testDuration(self):
        self.assertEqual(self.table.duration(), 2761413)


class MusicBrainzTestCase(tcommon.TestCase):
    # example taken from https://musicbrainz.org/doc/Disc_ID_Calculation
    # disc is Ettella Diamant

    def setUp(self):
        self.table = table.Table()

        for i in range(6):
            self.table.tracks.append(table.Track(i + 1, audio=True))

        offsets = [0, 15213, 32164, 46442, 63264, 80339]
        t = self.table.tracks
        for i, offset in enumerate(offsets):
            t[i].index(1, absolute=offset)

        self.assertFalse(self.table.hasTOC())

        self.table.leadout = 95312

        self.assertTrue(self.table.hasTOC())

    def testMusicBrainz(self):
        self.assertEqual(self.table.getMusicBrainzDiscId(),
                         '49HHV7Eb8UKF3aQiNmu1GR8vKTY-')


class PregapTestCase(tcommon.TestCase):

    def setUp(self):
        self.table = table.Table()

        for i in range(2):
            self.table.tracks.append(table.Track(i + 1, audio=True))

        offsets = [0, 15537]
        t = self.table.tracks
        for i, offset in enumerate(offsets):
            t[i].index(1, absolute=offset)
        t[1].index(0, offsets[1] - 200)

    def testPreGap(self):
        self.assertEqual(self.table.tracks[0].getPregap(), 0)
        self.assertEqual(self.table.tracks[1].getPregap(), 200)



# -*- Mode: Python -*-
# vi:si:et:sw=4:sts=4:ts=4

import os
import signal
import subprocess

from whipper.extern import asyncsub
from whipper.extern.task import task

import logging
logger = logging.getLogger(__name__)


class SyncRunner(task.SyncRunner):
    pass


class LoggableTask(task.Task):
    pass


class LoggableMultiSeparateTask(task.MultiSeparateTask):
    pass


class PopenTask(task.Task):
    """Task that runs a command using Popen."""

    logCategory = 'PopenTask'
    bufsize = 1024
    command = None
    cwd = None

    def start(self, runner):
        task.Task.start(self, runner)

        try:
            self._popen = asyncsub.Popen(self.command,
                                         bufsize=self.bufsize,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         close_fds=True, cwd=self.cwd)
        except OSError as e:
            import errno
            if e.errno == errno.ENOENT:
                self.commandMissing()

            raise

        logger.debug('started %r with pid %d', self.command, self._popen.pid)

        self.schedule(1.0, self._read, runner)

    def _read(self, runner):
        try:
            read = False

            ret = self._popen.recv()

            if ret:
                logger.debug("read from stdout: %s", ret)
                self.readbytesout(ret)
                read = True

            ret = self._popen.recv_err()

            if ret:
                logger.debug("read from stderr: %s", ret)
                self.readbyteserr(ret)
                read = True

            # if we read anything, we might have more to read, so
            # reschedule immediately
            if read and self.runner:
                self.schedule(0.0, self._read, runner)
                return

            # if we didn't read anything, give the command more time to
            # produce output
            if self._popen.poll() is None and self.runner:
                # not finished yet
                self.schedule(1.0, self._read, runner)
                return

            self._done()
        # FIXME: catching too general exception (Exception)
        except Exception as e:
            logger.debug('exception during _read(): %s', e)
            self.setException(e)
            self.stop()

    def _done(self):
        assert self._popen.returncode is not None, "No returncode"

        if self._popen.returncode >= 0:
            logger.debug('return code was %d', self._popen.returncode)
        else:
            logger.debug('terminated with signal %d', -self._popen.returncode)

        self.setProgress(1.0)

        if self._popen.returncode != 0:
            self.failed()
        else:
            self.done()

        self.stop()
        return

    def abort(self):
        logger.debug('aborting, sending SIGTERM to %d', self._popen.pid)
        os.kill(self._popen.pid, signal.SIGTERM)
        # self.stop()

    def readbytesout(self, bytes_stdout):
        """Call when bytes have been read from stdout."""
        pass

    def readbyteserr(self, bytes_stderr):
        """Call when bytes have been read from stderr."""
        pass

    def done(self):
        """Call when the command completed successfully."""
        pass

    def failed(self):
        """Call when the command failed."""
        pass

    def commandMissing(self):
        """Call when the command is missing."""
        pass



# -*- Mode: Python; test-case-name: whipper.test.test_program_cdparanoia -*-
# vi:si:et:sw=4:sts=4:ts=4

from whipper.program import cdrdao
from whipper.test import common

# TODO: Current test architecture makes testing cdrdao difficult. Revisit.


class VersionTestCase(common.TestCase):
    def testGetVersion(self):
        v = cdrdao.version()
        self.assertTrue(v)
        # make sure it starts with a digit
        self.assertTrue(int(v[0]))



import os

from whipper.common import common
from whipper.common import task as ctask

import logging
logger = logging.getLogger(__name__)

SOXI = 'soxi'


class AudioLengthTask(ctask.PopenTask):
    """
    Calculate the length of a track in audio samples.

    :cvar length: length of the decoded audio file, in audio samples
    :vartype length: int
    """

    logCategory = 'AudioLengthTask'
    description = 'Getting length of audio track'
    length = None

    def __init__(self, path):
        """
        Init AudioLengthTask.

        :param path: path to audio track
        :type path: str
        """
        assert isinstance(path, str), "%r is not str" % path

        self.logName = os.path.basename(path)

        self.command = [SOXI, '-s', path]

        self._error = []
        self._output = []

    def commandMissing(self):
        raise common.MissingDependencyException('soxi')

    def readbytesout(self, bytes_stdout):
        self._output.append(bytes_stdout)

    def readbyteserr(self, bytes_stderr):
        self._error.append(bytes_stderr)

    def failed(self):
        self.setException(Exception("soxi failed: %s" % "".join(self._error)))

    def done(self):
        if self._error:
            logger.warning("soxi reported on stderr: %s", "".join(self._error))
        self.length = int("".join(o.decode() for o in self._output))



import os
from subprocess import Popen, PIPE

import logging
logger = logging.getLogger(__name__)

SOX = 'sox'


def peak_level(track_path):
    """
    Accept a path to a sox-decodable audio file.

    :returns: track peak level from sox ('maximum amplitude')
    :rtype: float or None
    """
    if not os.path.exists(track_path):
        logger.warning("SoX peak detection failed: file not found")
        return None
    sox = Popen([SOX, track_path, "-n", "stats", "-b", "16"], stderr=PIPE)
    _, err = sox.communicate()
    if sox.returncode:
        logger.warning("SoX peak detection failed: %s", sox.returncode)
        return None
    # relevant captured lines looks like this:
    # Min level     -26215
    # Max level      26215
    min_level = int(err.splitlines()[2].split()[2])
    max_level = int(err.splitlines()[3].split()[2])
    return max(abs(min_level), abs(max_level))



from ruamel.yaml import YAML as ruamel_YAML
from ruamel.yaml.compat import StringIO

# https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string
class YAML(ruamel_YAML):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.width = 4000
        self.default_flow_style = False

    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        ruamel_YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()



# -*- Mode: Python -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

from whipper.command.basecommand import BaseCommand
from whipper.common import accurip, config, program
from whipper.extern.task import task
from whipper.image import image
from whipper.result import result

import logging
logger = logging.getLogger(__name__)


class Verify(BaseCommand):
    summary = "verify image"
    description = """
Verifies the image from the given .cue files against the AccurateRip database.
"""

    def add_arguments(self):
        self.parser.add_argument('cuefile', nargs='+', action='store',
                                 help="cue file to load rip image from")

    def do(self):
        prog = program.Program(config.Config())
        runner = task.SyncRunner()

        for arg in self.options.cuefile:
            cueImage = image.Image(arg)
            cueImage.setup(runner)

            # FIXME: this feels like we're poking at internals.
            prog.cuePath = arg
            prog.result = result.RipResult()
            for track in cueImage.table.tracks:
                tr = result.TrackResult()
                tr.number = track.number
                prog.result.tracks.append(tr)

            verified = False
            try:
                verified = prog.verifyImage(runner, cueImage.table)
            except accurip.EntryNotFound:
                print('AccurateRip entry not found')
            accurip.print_report(prog.result)
            if not verified:
                raise SystemExit(1)


class Image(BaseCommand):
    summary = "handle images"
    description = """
Handle disc images. Disc images are described by a .cue file.
Disc images can be verified.
"""
    subcommands = {
        'verify': Verify,
    }



# -*- Mode: Python; test-case-name: whipper.test.test_common_directory -*-
# vi:si:et:sw=4:sts=4:ts=4

from os.path import dirname, expanduser
from whipper.common import directory
from whipper.test import common


class DirectoryTestCase(common.TestCase):
    HOME = expanduser('~')
    HOME_PARENT = dirname(HOME)

    def testAll(self):
        path = directory.config_path()
        self.assertTrue(path.startswith(DirectoryTestCase.HOME_PARENT))



# -*- Mode: Python; test-case-name: whipper.test.test_common_drive -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

import os
from fcntl import ioctl

import logging
logger = logging.getLogger(__name__)


def _listify(listOrString):
    if isinstance(listOrString, str):
        return [listOrString, ]

    return listOrString


def getAllDevicePaths():
    try:
        # see https://savannah.gnu.org/bugs/index.php?38477
        return [str(dev) for dev in _getAllDevicePathsPyCdio()]
    except ImportError:
        logger.info('cannot import pycdio')
        return _getAllDevicePathsStatic()


def _getAllDevicePathsPyCdio():
    import pycdio
    import cdio

    # using FS_AUDIO here only makes it list the drive when an audio cd
    # is inserted
    # ticket 102: this cdio call returns a list of str, or a single str
    return _listify(cdio.get_devices_with_cap(pycdio.FS_MATCH_ALL, False))


def _getAllDevicePathsStatic():
    ret = []

    for c in ['/dev/cdrom', '/dev/cdrecorder']:
        if os.path.exists(c):
            ret.append(c)

    return ret


def getDeviceInfo(path):
    try:
        import cdio
    except ImportError:
        return None
    device = cdio.Device(path)
    _, vendor, model, release = device.get_hwinfo()

    return vendor, model, release


def get_cdrom_drive_status(drive_path):
    """
    Get the status of the disc drive.

    Drive status possibilities returned by CDROM_DRIVE_STATUS ioctl:
    - CDS_NO_INFO         = 0  (if not implemented)
    - CDS_NO_DISC         = 1
    - CDS_TRAY_OPEN       = 2
    - CDS_DRIVE_NOT_READY = 3
    - CDS_DISC_OK         = 4

    Documentation here:
    - https://www.kernel.org/doc/Documentation/ioctl/cdrom.txt
    - https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/include/uapi/linux/cdrom.h  # noqa: E501

    :param drive_path: path to the disc drive
    :type device: str
    :returns: return code of the 'CDROM_DRIVE_STATUS' ioctl
    :rtype: int
    """
    fd = os.open(drive_path, os.O_RDONLY | os.O_NONBLOCK)
    rc = ioctl(fd, 0x5326)  # AKA 'CDROM_DRIVE_STATUS'
    os.close(fd)
    return rc



# -*- Mode: Python; test-case-name: whipper.test.test_common_path -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

import re


class PathFilter:
    """Filter path components for safe storage on file systems."""

    def __init__(self, dot=True, posix=True, vfat=False, whitespace=False,
                 printable=False):
        """
        Init PathFilter.

        :param dot:        whether to strip leading dot
        :param posix:      whether to strip illegal chars in *nix OSes
        :param vfat:       whether to strip illegal chars in VFAT filesystems
        :param whitespace: whether to strip all whitespace chars
        :param printable:  whether to strip all non printable ASCII chars
        """
        self._dot = dot
        self._posix = posix
        self._vfat = vfat
        self._whitespace = whitespace
        self._printable = printable

    def filter(self, path):
        R_CH = '_'
        if self._dot:
            if path[:1] == '.':  # Slicing tolerant to empty strings
                path = R_CH + path[1:]
        if self._posix:
            path = re.sub(r'[\/\x00]', R_CH, path)
        if self._vfat:
            path = re.sub(r'[\x00-\x1F\x7F\"\*\/\:\<\>\?\\\|]', R_CH, path)
        if self._whitespace:
            path = re.sub(r'\s', R_CH, path)
        if self._printable:
            path = re.sub(r'[^\x20-\x7E]', R_CH, path)
        return path



# -*- Mode: Python; test-case-name: whipper.test.test_common_directory -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2013 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

from os import getenv, makedirs
from os.path import join, expanduser


def config_path():
    path = join(getenv('XDG_CONFIG_HOME') or join(expanduser('~'), '.config'),
                'whipper')
    makedirs(path, exist_ok=True)
    return join(path, 'whipper.conf')


def data_path(name=None):
    path = join(getenv('XDG_DATA_HOME') or
                join(expanduser('~'), '.local/share'),
                'whipper')
    if name:
        path = join(path, name)
    makedirs(path, exist_ok=True)
    return path



# -*- Mode: Python; test-case-name: whipper.test.test_common_encode -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.


from mutagen.flac import FLAC, Picture
from mutagen.id3 import PictureType

from whipper.extern.task import task

from whipper.program import sox
from whipper.program import flac

import logging
logger = logging.getLogger(__name__)


class SoxPeakTask(task.Task):
    description = 'Calculating peak level'

    def __init__(self, track_path):
        self.track_path = track_path
        self.peak = None

    def start(self, runner):
        task.Task.start(self, runner)
        self.schedule(0.0, self._sox_peak)

    def _sox_peak(self):
        self.peak = sox.peak_level(self.track_path)
        self.stop()


class FlacEncodeTask(task.Task):
    description = 'Encoding to FLAC'

    def __init__(self, track_path, track_out_path, what="track"):
        self.track_path = track_path
        self.track_out_path = track_out_path
        self.new_path = None
        self.description = 'Encoding %s to FLAC' % what

    def start(self, runner):
        task.Task.start(self, runner)
        self.schedule(0.0, self._flac_encode)

    def _flac_encode(self):
        flac.encode(self.track_path, self.track_out_path)
        self.stop()


class TaggingTask(task.Task):
    # TODO: Wizzup: Do we really want this as 'Task'...?
    # I only made it a task for now because that it's easier to integrate in
    # program/cdparanoia.py - where whipper currently does the tagging.
    # We should just move the tagging to a more sensible place.

    description = 'Writing tags to FLAC'

    def __init__(self, track_path, tags):
        self.track_path = track_path
        self.tags = tags

    def start(self, runner):
        task.Task.start(self, runner)
        self.schedule(0.0, self._tag)

    def _tag(self):
        w = FLAC(self.track_path)

        for k, v in list(self.tags.items()):
            w[k] = v

        w.save()

        self.stop()


class EmbedPictureTask(task.Task):
    description = 'Embed picture to FLAC'

    def __init__(self, track_path, cover_art_path):
        self.track_path = track_path
        self.cover_art_path = cover_art_path

    def start(self, runner):
        task.Task.start(self, runner)
        self.schedule(0.0, self._embed_picture)

    def _make_flac_picture(self, cover_art_filename):
        """
        Given a path to a jpg/png file, return a FLAC picture for embedding.

        The embedding will be performed using the mutagen module.

        :param cover_art_filename: path to cover art image file
        :type cover_art_filename: str
        :returns: a valid FLAC picture for embedding
        :rtype: mutagen.flac.Picture or None
        """
        if not cover_art_filename:
            return

        from PIL import Image

        im = Image.open(cover_art_filename)
        # NOTE: the cover art thumbnails we're getting from the Cover Art
        # Archive should be always in the JPEG format: this check is currently
        # useless but will leave it here to better handle unexpected formats.
        if im.format == 'JPEG':
            mime = 'image/jpeg'
        elif im.format == 'PNG':
            mime = 'image/png'
        else:
            # we only support png and jpeg
            logger.warning("no cover art will be added because the fetched "
                           "image format is unsupported")
            return

        pic = Picture()
        with open(cover_art_filename, 'rb') as f:
            pic.data = f.read()

        pic.type = PictureType.COVER_FRONT
        pic.mime = mime
        pic.width, pic.height = im.size
        if im.mode not in ('P', 'RGB', 'SRGB'):
            logger.warning("no cover art will be added because the fetched "
                           "image mode is unsupported")
            return

        return pic

    def _embed_picture(self):
        """
        Get flac picture generated from mutagen.flac.Picture then embed
        it to given track if the flac picture exists.
        """
        flac_pic = self._make_flac_picture(self.cover_art_path)
        if flac_pic:
            w = FLAC(self.track_path)
            w.add_picture(flac_pic)

        self.stop()



# -*- Mode: Python -*-
# vi:si:et:sw=4:sts=4:ts=4

import os

from whipper.command.main import main


if __name__ == '__main__':
    # Make accuraterip_checksum be found automatically if it was built
    local_arb = os.path.join(os.path.dirname(__file__), '..', 'src')
    os.environ['PATH'] = ':'.join([os.getenv('PATH'), local_arb])
    raise SystemExit(main())



# -*- Mode: Python; test-case-name: whipper.test.test_common_accurip -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2017 Samantha Baldwin
# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

import struct
import whipper
import os
from urllib.error import URLError, HTTPError
from urllib.request import urlopen, Request

from whipper.program.arc import accuraterip_checksum

import logging
logger = logging.getLogger(__name__)


ACCURATERIP_URL = "http://www.accuraterip.com/accuraterip/"


class EntryNotFound(Exception):
    pass


class _AccurateRipResponse:
    """
    An AR resp. contains a collection of metadata identifying a specific disc.

    For disc level metadata it contains the track count, two internal disc
    IDs, and the CDDB disc ID.

    A checksum and a confidence score is stored sequentially for each track in
    the disc index, which excludes any audio hidden in track pre-gaps (such as
    HTOA).

    The response is stored as a packed binary structure.
    """

    def __init__(self, data):
        """
        Init _AccurateRipResponse.

        Checksums and confidences arrays are indexed by relative track
        position, so track 1 will have array index 0, track 2 will have array
        index 1, and so forth. HTOA and other hidden tracks are not included.
        """
        self.num_tracks = data[0]
        self.discId1 = "%08x" % struct.unpack("<L", data[1:5])[0]
        self.discId2 = "%08x" % struct.unpack("<L", data[5:9])[0]
        self.cddbDiscId = "%08x" % struct.unpack("<L", data[9:13])[0]

        self.confidences = []
        self.checksums = []
        pos = 13
        for _ in range(self.num_tracks):
            confidence = data[pos]
            checksum = "%08x" % struct.unpack("<L", data[pos + 1:pos + 5])[0]
            self.confidences.append(confidence)
            self.checksums.append(checksum)
            pos += 9

    def __eq__(self, other):
        return [
            self.num_tracks, self.discId1, self.discId2, self.cddbDiscId,
            self.confidences, self.checksums
        ] == [
            other.num_tracks, other.discId1, other.discId2, other.cddbDiscId,
            other.confidences, other.checksums
        ]


def _split_responses(raw_entry):
    responses = []
    while raw_entry:
        track_count = raw_entry[0]
        nbytes = 1 + 12 + track_count * (1 + 8)
        responses.append(_AccurateRipResponse(raw_entry[:nbytes]))
        raw_entry = raw_entry[nbytes:]
    return responses


def calculate_checksums(track_paths):
    """
    Calculate AccurateRip checksums for the given tracks.

    HTOA checksums are not included in the database and are not calculated.

    :returns: ARv1 and ARv2 checksums as two arrays of character strings in a
              dictionary: ``{'v1': ['deadbeef', ...], 'v2': [...]}``
              or None instead of checksum string for unchecksummable tracks.
    :rtype: dict(string, list()) or None
    """
    track_count = len(track_paths)
    v1_checksums = []
    v2_checksums = []
    logger.debug('checksumming %d tracks', track_count)
    # This is done sequentially because it is very fast.
    for i, path in enumerate(track_paths):
        if os.path.exists(path):
            v1_sum, v2_sum = accuraterip_checksum(path, i+1, track_count)
        else:
            logger.warning('Can\'t checksum %s; path doesn\'t exist', path)
            v1_sum, v2_sum = None, None
        if v1_sum is None:
            logger.error('could not calculate AccurateRip v1 checksum '
                         'for track %d %r', i + 1, path)
            v1_checksums.append(None)
        else:
            v1_checksums.append("%08x" % v1_sum)
        if v2_sum is None:
            logger.error('could not calculate AccurateRip v2 checksum '
                         'for track %d %r', i + 1, path)
            v2_checksums.append(None)
        else:
            v2_checksums.append("%08x" % v2_sum)
    return {'v1': v1_checksums, 'v2': v2_checksums}


def _download_entry(path):
    url = ACCURATERIP_URL + path
    UA = "whipper/%s ( https://github.com/whipper-team/whipper )" % whipper.__version__  # noqa: E501
    logger.debug('downloading AccurateRip entry from %s', url)
    try:
        with urlopen(Request(url, headers={'User-Agent': UA})) as resp:
            return resp.read()
    except (URLError, HTTPError) as e:
        logger.error('error retrieving AccurateRip entry: %s', e)


def get_db_entry(path):
    """
    Download entry from accuraterip.com.

    ``path`` is in the format of the output of ``table.accuraterip_path()``.
    """
    raw_entry = _download_entry(path)
    if not raw_entry:
        logger.warning('entry not found in AccurateRip database')
        raise EntryNotFound
    return _split_responses(raw_entry)


def _assign_checksums_and_confidences(tracks, checksums, responses):
    for i, track in enumerate(tracks):
        for v in ('v1', 'v2'):
            track.AR[v]['CRC'] = checksums[v][i]
        track.AR['DBMaxConfidence'], track.AR['DBMaxConfidenceCRC'] = max(
            [(r.confidences[i], r.checksums[i]) for r in responses],
            key=lambda t: t[0]
        )


def _match_responses(tracks, responses):
    """
    Match and save track AR response checksums against all non-hidden tracks.

    :returns: True if every track has a match for every entry for either
              AccurateRip version, False otherwise.
    :rtype: bool
    """
    for r in responses:
        for i, track in enumerate(tracks):
            for v in ('v1', 'v2'):
                if track.AR[v]['CRC'] == r.checksums[i]:
                    if (track.AR[v]['DBConfidence'] is None or
                            r.confidences[i] > track.AR[v]['DBConfidence']):
                        track.AR[v]['DBCRC'] = r.checksums[i]
                        track.AR[v]['DBConfidence'] = r.confidences[i]
                        logger.debug(
                            'track %d matched response %s in AccurateRip'
                            ' database: %s crc %s confidence %s',
                            i, r.cddbDiscId, v, track.AR[v]['DBCRC'],
                            track.AR[v]['DBConfidence'])
    return any((
        all([t.AR['v1']['DBCRC'] for t in tracks]),
        all([t.AR['v2']['DBCRC'] for t in tracks])
    ))


def verify_result(result, responses, checksums):
    """
    Verify track AccurateRip checksums against database responses.

    Store track checksums and database values on result.
    """
    if not (result and responses and checksums):
        return False
    # exclude HTOA from AccurateRip verification
    # NOTE: if pre-gap hidden audio support is expanded to include
    # tracks other than HTOA, this is invalid.
    tracks = [t for t in result.tracks if t.number != 0]
    if not tracks:
        return False
    _assign_checksums_and_confidences(tracks, checksums, responses)
    return _match_responses(tracks, responses)


def print_report(result):
    """Print AccurateRip verification results."""
    for _, track in enumerate(result.tracks):
        status = 'rip NOT accurate'
        conf = '(not found)'
        db = 'notfound'
        if track.AR['DBMaxConfidence'] is not None:
            db = track.AR['DBMaxConfidenceCRC']
            conf = '(max confidence    %3d)' % track.AR['DBMaxConfidence']
            if track.AR['v1']['DBCRC'] or track.AR['v2']['DBCRC']:
                status = 'rip accurate'
                db = ', '.join([_f for _f in (
                    track.AR['v1']['DBCRC'],
                    track.AR['v2']['DBCRC']
                ) if _f])
            max_conf = max(
                [track.AR[v]['DBConfidence'] for v in ('v1', 'v2')
                 if track.AR[v]['DBConfidence'] is not None], default=None
            )
            if max_conf:
                if max_conf < track.AR['DBMaxConfidence']:
                    conf = '(confidence %3d of %3d)' % (
                        max_conf, track.AR['DBMaxConfidence']
                    )
        # htoa tracks (i == 0) do not have an ARCRC
        if track.number == 0:
            print('track  0: unknown          (not tracked)')
            continue
        if not (track.AR['v1']['CRC'] or track.AR['v2']['CRC']):
            logger.error('no track AR CRC on non-HTOA track %d', track.number)
            print('track %2d: unknown          (error)' % track.number)
        else:
            print('track %2d: %-16s %-23s v1 [%s], v2 [%s], DB [%s]' % (
                track.number, status, conf,
                track.AR['v1']['CRC'], track.AR['v2']['CRC'], db
            ))



# -*- Mode: Python; test-case-name: whipper.test.test_image_toc -*-
# vi:si:et:sw=4:sts=4:ts=4

import os
import copy
import shutil
import tempfile

from whipper.image import toc

from whipper.test import common


class CureTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'cure.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 13)

    def testGetTrackLength(self):
        t = self.toc.table.tracks[0]
        # first track has known length because the .toc is a single file
        # its length is all of track 1 from .toc, plus the INDEX 00 length
        # of track 2
        self.assertEqual(self.toc.getTrackLength(t),
                         (((6 * 60) + 16) * 75 + 45) + ((1 * 75) + 4))
        # last track has unknown length
        t = self.toc.table.tracks[-1]
        self.assertEqual(self.toc.getTrackLength(t), -1)

    def testIndexes(self):
        # track 2, index 0 is at 06:16:45 or 28245
        # track 2, index 1 is at 06:17:49 or 28324
        # FIXME: cdrdao seems to get length of FILE 1 frame too many,
        # and START value one frame less
        t = self.toc.table.tracks[1]
        self.assertEqual(t.getIndex(0).relative, 28245)
        self.assertEqual(t.getIndex(1).relative, 28324)

    def _getIndex(self, t, i):
        track = self.toc.table.tracks[t - 1]
        return track.getIndex(i)

    def _assertAbsolute(self, t, i, value):
        index = self._getIndex(t, i)
        self.assertEqual(index.absolute, value)

    def _assertPath(self, t, i, value):
        index = self._getIndex(t, i)
        self.assertEqual(index.path, value)

    def _assertRelative(self, t, i, value):
        index = self._getIndex(t, i)
        self.assertEqual(index.relative, value)

    def testSetFile(self):
        self._assertAbsolute(1, 1, 0)
        self._assertAbsolute(2, 0, 28245)
        self._assertAbsolute(2, 1, 28324)
        self._assertPath(1, 1, "data.wav")

        self.toc.table.clearFiles()

        self._assertAbsolute(1, 1, 0)
        self._assertAbsolute(2, 0, 28245)
        self._assertAbsolute(2, 1, 28324)
        self._assertAbsolute(3, 1, 46110)
        self._assertAbsolute(4, 1, 66767)
        self._assertPath(1, 1, None)
        self._assertRelative(1, 1, None)

        # adding the first track file with length 28324 to the table should
        # relativize from absolute 0 to absolute 28323, right before track 2,
        # index 1
        self.toc.table.setFile(1, 1, 'track01.wav', 28324)
        self._assertPath(1, 1, 'track01.wav')
        self._assertRelative(1, 1, 0)
        self._assertPath(2, 0, 'track01.wav')
        self._assertRelative(2, 0, 28245)

        self._assertPath(2, 1, None)
        self._assertRelative(2, 1, None)

    def testConvertCue(self):
        cue = self.toc.table.cue()
        ref = self.readCue('cure.cue')
        common.diffStrings(ref, cue)

        # we verify it because it has failed in readdisc in the past
        self.assertEqual(self.toc.table.accuraterip_path(),
                         '3/c/4/dBAR-013-0019d4c3-00fe8924-b90c650d.bin')

    def testGetRealPath(self):
        self.assertRaises(KeyError, self.toc.getRealPath, 'track01.wav')
        (fd, path) = tempfile.mkstemp(suffix='.whipper.test.wav')
        self.assertEqual(self.toc.getRealPath(path), path)

        winpath = path.replace('/', '\\')
        self.assertEqual(self.toc.getRealPath(winpath), path)
        os.close(fd)
        os.unlink(path)

# Bloc Party - Silent Alarm has a Hidden Track One Audio


class BlocTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'bloc.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 13)

    def testGetTrackLength(self):
        t = self.toc.table.tracks[0]
        # first track has known length because the .toc is a single file
        # the length is from Track 1, Index 1 to Track 2, Index 1, so
        # does not include the htoa
        self.assertEqual(self.toc.getTrackLength(t), 19649)
        # last track has unknown length
        t = self.toc.table.tracks[-1]
        self.assertEqual(self.toc.getTrackLength(t), -1)

    def testIndexes(self):
        track01 = self.toc.table.tracks[0]
        index00 = track01.getIndex(0)
        self.assertEqual(index00.absolute, 0)
        self.assertEqual(index00.relative, 0)
        self.assertEqual(index00.counter, 0)

        index01 = track01.getIndex(1)
        self.assertEqual(index01.absolute, 15220)
        self.assertEqual(index01.relative, 0)
        self.assertEqual(index01.counter, 1)

        track05 = self.toc.table.tracks[4]

        index00 = track05.getIndex(0)
        self.assertEqual(index00.absolute, 84070)
        self.assertEqual(index00.relative, 68850)
        self.assertEqual(index00.counter, 1)

        index01 = track05.getIndex(1)
        self.assertEqual(index01.absolute, 84142)
        self.assertEqual(index01.relative, 68922)
        self.assertEqual(index01.counter, 1)

    # This disc has a pre-gap, so is a good test for .CUE writing

    def testConvertCue(self):
        self.assertTrue(self.toc.table.hasTOC())
        cue = self.toc.table.cue()
        ref = self.readCue('bloc.cue')
        common.diffStrings(ref, cue)

    def testCDDBId(self):
        # cd-discid output:
        # ad0be00d 13 15370 35019 51532 69190 84292 96826 112527 132448
        # 148595 168072 185539 203331 222103 3244

        self.assertEqual(self.toc.table.getCDDBDiscId(), 'ad0be00d')

    def testAccurateRip(self):
        # we verify it because it has failed in readdisc in the past
        self.assertEqual(self.toc.table.accuraterip_path(),
                         'e/d/2/dBAR-013-001af2de-0105994e-ad0be00d.bin')

# The Breeders - Mountain Battles has CDText


class BreedersTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'breeders.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 13)

    def testCDText(self):
        cdt = self.toc.table.cdtext
        self.assertEqual(cdt['PERFORMER'], 'THE BREEDERS')
        self.assertEqual(cdt['TITLE'], 'MOUNTAIN BATTLES')

        t = self.toc.table.tracks[0]
        cdt = t.cdtext
        self.assertRaises(AttributeError, getattr, cdt, 'PERFORMER')
        self.assertEqual(cdt['TITLE'], 'OVERGLAZED')

    def testConvertCue(self):
        self.assertTrue(self.toc.table.hasTOC())
        cue = self.toc.table.cue()
        ref = self.readCue('breeders.cue')
        self.assertEqual(cue, ref)

# Ladyhawke has a data track


class LadyhawkeTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'ladyhawke.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 13)
        self.assertFalse(self.toc.table.tracks[-1].audio)

    def testCDDBId(self):
        self.assertEqual(self.toc.table.getCDDBDiscId(), 'c60af50d')
        # output from cd-discid:
        # c60af50d 13 150 15687 31841 51016 66616 81352 99559 116070 133243
        # 149997 161710 177832 207256 2807

    def testMusicBrainz(self):
        self.assertEqual(self.toc.table.getMusicBrainzDiscId(),
                         "KnpGsLhvH.lPrNc1PBL21lb9Bg4-")
        self.assertEqual(self.toc.table.getMusicBrainzSubmitURL(),
                         "https://musicbrainz.org/cdtoc/attach?toc=1+12+195856+150+15687+31841+51016+66616+81352+99559+116070+133243+149997+161710+177832&tracks=12&id=KnpGsLhvH.lPrNc1PBL21lb9Bg4-")  # noqa: E501

    # FIXME: I don't trust this toc, but I can't find the CD anymore

    def testDuration(self):
        self.assertEqual(self.toc.table.duration(), 2761413)

    def testGetFrameLength(self):
        self.assertEqual(self.toc.table.getFrameLength(data=True), 210385)

    def testCue(self):
        self.assertTrue(self.toc.table.canCue())
        data = self.toc.table.cue()
        lines = data.split("\n")
        self.assertEqual(lines[0], "REM DISCID C60AF50D")


class CapitalMergeTestCase(common.TestCase):

    def setUp(self):
        self.toc1 = toc.TocFile(os.path.join(os.path.dirname(__file__),
                                             'capital.1.toc'))
        self.toc1.parse()
        self.assertEqual(len(self.toc1.table.tracks), 11)
        self.assertTrue(self.toc1.table.tracks[-1].audio)

        self.toc2 = toc.TocFile(os.path.join(os.path.dirname(__file__),
                                             'capital.2.toc'))
        self.toc2.parse()
        self.assertEqual(len(self.toc2.table.tracks), 1)
        self.assertFalse(self.toc2.table.tracks[-1].audio)

        self.table = copy.deepcopy(self.toc1.table)
        self.table.merge(self.toc2.table)

    def testCDDBId(self):
        self.assertEqual(self.table.getCDDBDiscId(), 'b910140c')
        # output from cd-discid:
        # b910140c 12 24320 44855 64090 77885 88095 104020 118245 129255 141765
        # 164487 181780 209250 4440

    def testMusicBrainz(self):
        # URL to submit: https://musicbrainz.org/cdtoc/attach?toc=1+11+
        # 197850+24320+44855+64090+77885+88095+104020+118245+129255+141765+
        # 164487+181780&tracks=11&id=MAj3xXf6QMy7G.BIFOyHyq4MySE-
        self.assertEqual(self.table.getMusicBrainzDiscId(),
                         "MAj3xXf6QMy7G.BIFOyHyq4MySE-")

    def testDuration(self):
        # this matches track 11 end sector - track 1 start sector on
        # MusicBrainz
        # compare to 3rd and 4th value in URL above
        self.assertEqual(self.table.getFrameLength(), 173530)
        self.assertEqual(self.table.duration(), 2313733)

    def testMusicBrainzDataTrackFirst(self):
        self.table = copy.deepcopy(self.toc2.table)
        self.table.merge(self.toc1.table)
        print(self.table.tracks)
        self.assertEqual(self.table.getMusicBrainzDiscId(),
                         "QTYYFFAgNK4Np2EHjfPTBavqtw8-")


class UnicodeTestCase(common.TestCase, common.UnicodeTestMixin):

    def setUp(self):
        # we copy the normal non-utf8 filename to a utf-8 filename
        # in this test because builds with LANG=C fail if we include
        # utf-8 filenames in the dist
        path = 'Jos\xe9Gonz\xe1lez.toc'
        self._performer = 'Jos\xe9 Gonz\xe1lez'
        source = os.path.join(os.path.dirname(__file__), 'jose.toc')
        (fd, self.dest) = tempfile.mkstemp(suffix=path)
        os.close(fd)
        shutil.copy(source, self.dest)
        self.toc = toc.TocFile(self.dest)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 10)

    def tearDown(self):
        os.unlink(self.dest)

    def testGetTrackLength(self):
        t = self.toc.table.tracks[0]
        # first track has known length because the .toc is a single file
        self.assertEqual(self.toc.getTrackLength(t), 12001)
        # last track has unknown length
        t = self.toc.table.tracks[-1]
        self.assertEqual(self.toc.getTrackLength(t), -1)

    def testGetTrackPerformer(self):
        t = self.toc.table.tracks[0]
        self.assertEqual(t.cdtext['PERFORMER'], self._performer)


# Interpol - Turn of the Bright Lights has same cddb disc id as
# Afghan Whigs - Gentlemen


class TOTBLTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'totbl.fast.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 11)

    def testCDDBId(self):
        self.assertEqual(self.toc.table.getCDDBDiscId(), '810b7b0b')


class GentlemenTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__),
                                 'gentlemen.fast.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEquals(len(self.toc.table.tracks), 11)

    def testCDDBId(self):
        self.toc.table.absolutize()
        self.assertEquals(self.toc.table.getCDDBDiscId(), '810b7b0b')


# The Strokes - Someday has a 1 frame SILENCE marked as such in toc


class StrokesTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__),
                                 'strokes-someday.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 1)

    def testIndexes(self):
        t = self.toc.table.tracks[0]
        i0 = t.getIndex(0)
        self.assertEqual(i0.relative, 0)
        self.assertEqual(i0.absolute, 0)
        self.assertEqual(i0.counter, 0)
        self.assertEqual(i0.path, None)

        i1 = t.getIndex(1)
        self.assertEqual(i1.relative, 0)
        self.assertEqual(i1.absolute, 1)
        self.assertEqual(i1.counter, 1)
        self.assertEqual(i1.path, 'data.wav')

        cue = self._filterCue(self.toc.table.cue())
        with open(os.path.join(os.path.dirname(__file__),
                               'strokes-someday.eac.cue')) as f:
            ref = self._filterCue(f.read())
        common.diffStrings(ref, cue)

    @staticmethod
    def _filterCue(output):
        # helper to be able to compare our generated .cue with the
        # EAC-extracted one
        discard = ['TITLE', 'PERFORMER', 'FLAGS', 'REM']
        lines = output.split('\n')

        res = []

        for line in lines:
            found = False
            for needle in discard:
                if line.find(needle) > -1:
                    found = True

            if line.find('FILE') > -1:
                line = 'FILE "data.wav" WAVE'

            if not found:
                res.append(line)

        return '\n'.join(res)


# Surfer Rosa has
# track 00 consisting of 32 frames of SILENCE
# track 11 Vamos with an INDEX 02
# compared to an EAC single .cue file, all our offsets are 32 frames off
# because the toc uses silence for track 01 index 00 while EAC puts it in
# Range.wav
class SurferRosaTestCase(common.TestCase):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__), 'surferrosa.toc')
        self.toc = toc.TocFile(self.path)
        self.toc.parse()
        self.assertEqual(len(self.toc.table.tracks), 21)

    def testIndexes(self):
        # HTOA
        t = self.toc.table.tracks[0]
        self.assertEqual(len(t.indexes), 2)

        i0 = t.getIndex(0)
        self.assertEqual(i0.relative, 0)
        self.assertEqual(i0.absolute, 0)
        self.assertEqual(i0.path, None)
        self.assertEqual(i0.counter, 0)

        i1 = t.getIndex(1)
        self.assertEqual(i1.relative, 0)
        self.assertEqual(i1.absolute, 32)
        self.assertEqual(i1.path, 'data.wav')
        self.assertEqual(i1.counter, 1)

        # track 11, Vamos

        t = self.toc.table.tracks[10]
        self.assertEqual(len(t.indexes), 2)

        # 32 frames of silence, and 1483 seconds of data.wav
        self.assertEqual(t.getIndex(1).relative, 111225)
        self.assertEqual(t.getIndex(1).absolute, 111257)
        self.assertEqual(t.getIndex(2).relative, 111225 + 3370)
        self.assertEqual(t.getIndex(2).absolute, 111257 + 3370)



# -*- Mode: Python; test-case-name: whipper.test.test_image_table -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

"""Wrap Table of Contents."""

import copy
from urllib.parse import urlunparse, urlencode

import whipper

from whipper.common import common, config
from whipper.extern.freedb import DiscID

import logging
logger = logging.getLogger(__name__)

# FIXME: taken from libcdio, but no reference found for these

CDTEXT_FIELDS = [
    'ARRANGER',
    'COMPOSER',
    'DISCID',
    'GENRE',
    'MESSAGE',
    'ISRC',
    'PERFORMER',
    'SIZE_INFO',
    'SONGWRITER',
    'TITLE',
    'TOC_INFO',
    'TOC_INFO2',
    'UPC_EAN',
]


class Track:
    """
    Represent a track entry in a Table.

    :cvar number: track number (1-based)
    :vartype number: int
    :cvar audio: whether the track is audio
    :vartype audio: bool
    :cvar indexes: dict of number
    :vartype indexes: dict of number -> :any:`Index`
    :cvar isrc: ISRC code (12 alphanumeric characters)
    :vartype isrc: str
    :cvar cdtext: dictionary of CD Text information;
                  :any:`see CDTEXT_KEYS`
    :vartype cdtext: str
    :cvar pre_emphasis: whether track is pre-emphasised
    :vartype pre_emphasis: bool
    """

    number = None
    audio = None
    indexes = None
    isrc = None
    cdtext = None
    session = None
    pre_emphasis = None

    def __repr__(self):
        return '<Track %02d>' % self.number

    def __init__(self, number, audio=True, session=None):
        self.number = number
        self.audio = audio
        self.indexes = {}
        self.cdtext = {}

    def index(self, number, absolute=None, path=None, relative=None,
              counter=None):
        """
        Instantiate Index object and store it in class variable.

        :param number: index number
        :type number: int
        :param absolute: absolute index offset, in CD frames
        :type absolute: int or None
        :param path: path to track
        :type path: str or None
        :param relative: relative index offset, in CD frames
        :type relative: int or None
        :param counter: the source counter; updates for each different
                        data source (silence or different file path)
        :type counter: int or None
        """
        if path is not None:
            assert isinstance(path, str), "%r is not str" % path

        i = Index(number, absolute, path, relative, counter)
        self.indexes[number] = i

    def getIndex(self, number):
        return self.indexes[number]

    def getFirstIndex(self):
        """
        Get the first chronological index for this track.

        Typically this is INDEX 01; but it could be INDEX 00 if there's
        a pre-gap.
        """
        indexes = sorted(self.indexes.keys())
        return self.indexes[indexes[0]]

    def getLastIndex(self):
        indexes = sorted(self.indexes.keys())
        return self.indexes[indexes[-1]]

    def getPregap(self):
        """
        Return the length of the pregap for this track.

        The pregap is 0 if there is no index 0, and the difference between
        index 1 and index 0 if there is.
        """
        if 0 not in self.indexes:
            return 0

        return self.indexes[1].absolute - self.indexes[0].absolute


class Index:
    """
    Represent an index of a track on a CD.

    :cvar counter: counter for the index source; distinguishes between
                   the matching FILE lines in .cue files for example
    :vartype counter: int
    :cvar path: path to track
    :vartype path: str or None
    """

    number = None
    absolute = None
    path = None
    relative = None
    counter = None

    def __init__(self, number, absolute=None, path=None, relative=None,
                 counter=None):

        if path is not None:
            assert isinstance(path, str), "%r is not str" % path

        self.number = number
        self.absolute = absolute
        self.path = path
        self.relative = relative
        self.counter = counter

    def __repr__(self):
        return '<Index %02d absolute %r path %r relative %r counter %r>' % (
            self.number, self.absolute, self.path, self.relative, self.counter)


class Table:
    """
    Represent a table of indexes on a CD.

    :cvar tracks: tracks on this CD
    :vartype tracks: list(Track)
    :cvar catalog: catalog number
    :vartype catalog: str
    """

    tracks = None  # list of Track
    leadout = None  # offset where the leadout starts
    catalog = None  # catalog number; FIXME: is this UPC ?
    cdtext = None
    mbdiscid = None

    classVersion = 4

    def __init__(self, tracks=None):
        if not tracks:
            tracks = []

        self.tracks = tracks
        self.cdtext = {}
        # done this way because just having a class-defined instance var
        # gets overridden when unpickling
        self.instanceVersion = self.classVersion
        self.unpickled()

    def unpickled(self):
        self.logName = "Table 0x%08x v%d" % (id(self), self.instanceVersion)
        logger.debug('set logName')

    def getTrackStart(self, number):
        """
        Return the start of the given track number's index 1, in CD frames.

        :param number: the track number, 1-based
        :type number: int
        :returns: the start of the given track number's index 1, in CD frames
        :rtype: int
        """
        track = self.tracks[number - 1]
        return track.getIndex(1).absolute

    def getTrackEnd(self, number):
        """
        Return the end of the given track number, in CD frames.

        :param number: the track number, 1-based
        :type number: int
        :returns: the end of the given track number (ie index 1 of next track)
        :rtype: int
        """
        # default to end of disc
        end = self.leadout - 1

        # if not last track, calculate it from the next track
        if number < len(self.tracks):
            end = self.tracks[number].getIndex(1).absolute - 1

            # if on a session border, subtract the session leadin
            thisTrack = self.tracks[number - 1]
            nextTrack = self.tracks[number]
            # The session attribute of a track is None by default (session 1)
            # with value > 1 if the track is in another session. Py3 doesn't
            # allow NoneType comparisons so we compare against 1 in that case
            if int(nextTrack.session or 1) > int(thisTrack.session or 1):
                gap = self._getSessionGap(nextTrack.session)
                end -= gap

        return end

    def getTrackLength(self, number):
        """
        Return the length, in CD frames, for the given track number.

        :param number: the track number, 1-based
        :type number: int
        :returns: the length of the given track number, in CD frames
        :rtype: int
        """
        return self.getTrackEnd(number) - self.getTrackStart(number) + 1

    def getAudioTracks(self):
        """
        Return the number of audio tracks on the disc.

        :returns: the number of audio tracks on the disc
        :rtype: int
        """
        return len([t for t in self.tracks if t.audio])

    def hasDataTracks(self):
        """
        Return whether the disc contains data tracks.

        :returns: whether the disc contains data tracks
        :rtype: bool
        """
        return len([t for t in self.tracks if not t.audio]) > 0

    @staticmethod
    def _cddbSum(i):
        ret = 0
        while i > 0:
            ret += (i % 10)
            i /= 10

        return ret

    def getCDDBValues(self):
        """
        Get all CDDB values needed to calculate disc id and lookup URL.

        This includes:

        * CDDB disc id
        * number of audio tracks
        * offset of index 1 of each track
        * length of disc in seconds (including data track)

        :rtype: list(int)
        """
        offsets = []

        # cddb disc id takes into account data tracks
        # last byte is the number of tracks on the CD
        n = 0

        # CD's have a standard lead-in time of 2 seconds
        # which gets added for CDDB disc id's
        delta = 2 * common.FRAMES_PER_SECOND

        debug = [str(len(self.tracks))]
        for track in self.tracks:
            offset = self.getTrackStart(track.number) + delta
            offsets.append(offset)
            debug.append(str(offset))
            seconds = offset // common.FRAMES_PER_SECOND
            n += self._cddbSum(seconds)

        # the 'real' leadout, not offset by 150 frames
        last = self.tracks[-1]
        leadout = self.getTrackEnd(last.number) + 1
        logger.debug('leadout LBA: %d', leadout)

        # FIXME: we can't replace these calculations with the getFrameLength
        # call because the start and leadout in the algorithm get rounded
        # before making the difference
        startSeconds = self.getTrackStart(1) // common.FRAMES_PER_SECOND
        leadoutSeconds = leadout // common.FRAMES_PER_SECOND
        t = leadoutSeconds - startSeconds
        # durationFrames = self.getFrameLength(data=True)
        # duration = durationFrames / common.FRAMES_PER_SECOND
        # assert t == duration, "%r != %r" % (t, duration)

        debug.append(str(leadoutSeconds + 2))  # 2 is the 150 frame cddb offset

        result = DiscID(offsets, t, len(self.tracks), leadoutSeconds)
        value = int(result)

        # compare this debug line to cd-discid output
        logger.debug('cddb values: %r', result)

        logger.debug('cddb disc id debug: %s',
                     " ".join(["%08x" % value, ] + debug))

        return result

    def getCDDBDiscId(self):
        """
        Calculate the CDDB disc ID.

        :returns: the 8-character hexadecimal disc ID
        :rtype: str
        """
        values = self.getCDDBValues()
        return "%08x" % int(values)

    def getMusicBrainzDiscId(self):
        """
        Calculate the MusicBrainz disc ID.

        :returns: the 28-character base64-encoded disc ID
        :rtype: str
        """
        if self.mbdiscid:
            logger.debug('getMusicBrainzDiscId: returning cached %r',
                         self.mbdiscid)
            return self.mbdiscid

        from discid import put

        values = self._getMusicBrainzValues()

        disc = put(values[0], values[1], values[2], values[3:])
        logger.debug('getMusicBrainzDiscId: returning %r', disc.id)
        self.mbdiscid = disc.id
        return disc.id

    def getMusicBrainzSubmitURL(self):
        serv = config.Config().get_musicbrainz_server()

        discid = self.getMusicBrainzDiscId()
        values = self._getMusicBrainzValues()

        query = urlencode([
            ('toc', ' '.join([str(v) for v in values])),
            ('tracks', self.getAudioTracks()),
            ('id', discid),
        ])

        return urlunparse((
            serv['scheme'], serv['netloc'], '/cdtoc/attach', '', query, ''))

    def getFrameLength(self, data=False):
        """
        Get the length in frames (excluding HTOA).

        :param data: whether to include the data tracks in the length
        :type data: bool
        """
        # the 'real' leadout, not offset by 150 frames
        if data:
            last = self.tracks[-1]
        else:
            last = self.tracks[self.getAudioTracks() - 1]

        leadout = self.getTrackEnd(last.number) + 1
        logger.debug('leadout LBA: %d', leadout)
        durationFrames = leadout - self.getTrackStart(1)

        return durationFrames

    def duration(self):
        """Get the duration in ms for all audio tracks (excluding HTOA)."""
        return int(self.getFrameLength() * 1000.0 / common.FRAMES_PER_SECOND)

    def _getMusicBrainzValues(self):
        """
        Get all MusicBrainz values needed to calculate disc id and submit URL.

        This includes:

        * track number of first track
        * number of audio tracks
        * leadout of disc
        * offset of index 1 of each track

        :rtype: list(int)
        """
        # MusicBrainz disc id does not take into account data tracks

        result = []

        # number of first track
        result.append(1)

        # number of last audio track (default: number of audio tracks)
        lastTrack = self.getAudioTracks()
        result.append(lastTrack)

        dataTrackLast = False
        additional = 0
        offsets = []

        # offsets of tracks
        for i in range(0, len(self.tracks)):
            track = self.tracks[i]
            if not track.audio:
                # if the data track is not at the end
                if i < len(self.tracks) - 1:
                    additional += 1
                else:
                    # if the data track is last
                    dataTrackLast = True
                    sectors = self.tracks[-1].getIndex(1).absolute - 11400
                    # treat leadout offset as track 0 offset
                    sectors += 150
                continue
            offset = track.getIndex(1).absolute + 150
            offsets.append(offset)

        if not dataTrackLast:
            # the end of the last audio track, +1 since getTrackEnd returned
            # value is always down by 1 unit. Which means that's actually
            # offsets[-1] + getTrackLength(lastTrack).
            sectors = self.getTrackEnd(lastTrack + additional) + 1 + 150

        result.append(sectors)
        result.extend(offsets)

        logger.debug('MusicBrainz values: %r', result)
        return result

    def cue(self, cuePath='', program='whipper'):
        """
        Dump our internal representation to a .cue file content.

        :param cuePath: path to the cue file to be written. If empty,
                        will treat paths as if in current directory
        :type cuePath: unicode
        :param program: name of the program (ripping software)
        :type program: str
        :rtype: str
        """
        logger.debug('generating .cue for cuePath %r', cuePath)

        lines = []

        def writeFile(path):
            targetPath = common.getRelativePath(path, cuePath)
            line = 'FILE "%s" WAVE' % targetPath
            lines.append(line)
            logger.debug('writeFile: %r', line)

        # header
        main = ['PERFORMER', 'TITLE']

        for key in CDTEXT_FIELDS:
            if key not in main and key in self.cdtext:
                lines.append("    %s %s" % (key, self.cdtext[key]))

        assert self.hasTOC(), "Table does not represent a full CD TOC"
        lines.append('REM DISCID %s' % self.getCDDBDiscId().upper())
        lines.append('REM COMMENT "%s %s"' % (program, whipper.__version__))

        if self.catalog:
            lines.append("CATALOG %s" % self.catalog)

        for key in main:
            if key in self.cdtext:
                lines.append('%s "%s"' % (key, self.cdtext[key]))

        # FIXME:
        # - the first FILE statement goes before the first TRACK, even if
        #   there is a non-file-using PREGAP
        # - the following FILE statements come after the last INDEX that
        #   use that FILE; so before a next TRACK, PREGAP silence, ...

        # add the first FILE line; EAC always puts the first FILE
        # statement before TRACK 01 and any possible PRE-GAP
        firstTrack = self.tracks[0]
        index = firstTrack.getFirstIndex()
        indexOne = firstTrack.getIndex(1)
        counter = index.counter
        track = firstTrack

        while not index.path:
            t, i = self.getNextTrackIndex(track.number, index.number)
            track = self.tracks[t - 1]
            index = track.getIndex(i)
            counter = index.counter

        if index.path:
            logger.debug('counter %d, writeFile', counter)
            writeFile(index.path)

        for i, track in enumerate(self.tracks):
            logger.debug('track i %r, track %r', i, track)
            # FIXME: skip data tracks for now
            if not track.audio:
                continue

            indexes = sorted(track.indexes.keys())

            wroteTrack = False

            for number in indexes:
                index = track.indexes[number]
                logger.debug('index %r, %r', number, index)

                # any time the source counter changes to a higher value,
                # write a FILE statement
                # it has to be higher, because we can run into the HTOA
                # at counter 0 here
                if index.counter > counter:
                    if index.path:
                        logger.debug('counter %d, writeFile', counter)
                        writeFile(index.path)
                    logger.debug('setting counter to index.counter %r',
                                 index.counter)
                    counter = index.counter

                # any time we hit the first index, write a TRACK statement
                if not wroteTrack:
                    wroteTrack = True
                    line = "  TRACK %02d %s" % (i + 1, 'AUDIO')
                    lines.append(line)
                    logger.debug('%r', line)

                    for key in CDTEXT_FIELDS:
                        if key in track.cdtext:
                            lines.append('    %s "%s"' % (
                                key, track.cdtext[key]))

                    if track.isrc is not None:
                        lines.append("    ISRC %s" % track.isrc)

                    if track.pre_emphasis is not None:
                        lines.append("    FLAGS PRE")

                    # handle TRACK 01 INDEX 00 specially
                    if 0 in indexes:
                        index00 = track.indexes[0]
                        if i == 0:
                            # if we have a silent pre-gap, output it
                            if not index00.path:
                                length = indexOne.absolute - index00.absolute
                                lines.append("    PREGAP %s" %
                                             common.framesToMSF(length))
                                continue

                        # handle any other INDEX 00 after its TRACK
                        lines.append("    INDEX "
                                     "%02d %s" % (0, common.framesToMSF(
                                                        index00.relative)))

                if number > 0:
                    # index 00 is output after TRACK up above
                    lines.append("    INDEX %02d %s" % (number,
                                                        common.framesToMSF(
                                                            index.relative)))

        lines.append("")

        return "\n".join(lines)

    # methods that modify the table

    def clearFiles(self):
        """
        Clear all file backings.

        Resets indexes paths and relative offsets.
        """
        # FIXME: do a loop over track indexes better, with a pythonic
        # construct that allows you to do for t, i in ...
        t = self.tracks[0].number
        index = self.tracks[0].getFirstIndex()
        i = index.number

        logger.debug('clearing path')
        while True:
            track = self.tracks[t - 1]
            index = track.getIndex(i)
            logger.debug('clearing path on track %d, index %d', t, i)
            index.path = None
            index.relative = None
            try:
                t, i = self.getNextTrackIndex(t, i)
            except IndexError:
                break

    def setFile(self, track, index, path, length, counter=None):
        """
        Set the given file as the source from the given index on.

        Will loop over all indexes that fall within the given length,
        to adjust the path.

        Assumes all indexes have an absolute offset and will raise if not.

        :param track: track number, 1-based
        :type track: int
        :param index: index of the track
        :type index: int
        :param path: path to track
        :type path: unicode
        :param length: length of the given track, in CD frames
        :type length: int
        :param counter: counter for the index source; distinguishes between
                        the matching FILE lines in .cue files for example
        :type counter: int or None
        """
        logger.debug('setFile: track %d, index %d, path %r, length %r, '
                     'counter %r', track, index, path, length, counter)

        t = self.tracks[track - 1]
        i = t.indexes[index]
        start = i.absolute
        assert start is not None, "index %r is missing absolute offset" % i
        end = start + length - 1  # last sector that should come from this file

        # FIXME: check border conditions here, esp. wrt. toc's off-by-one bug
        while i.absolute <= end:
            i.path = path
            i.relative = i.absolute - start
            i.counter = counter
            logger.debug('setting path %r, relative %r on track %d, '
                         'index %d, counter %r', path, i.relative, track,
                         index, counter)
            try:
                track, index = self.getNextTrackIndex(track, index)
                t = self.tracks[track - 1]
                i = t.indexes[index]
            except IndexError:
                break

    def absolutize(self):
        """
        Calculate absolute offsets on indexes as much as possible.

        Only possible for as long as tracks draw from the same file.
        """
        t = self.tracks[0].number
        index = self.tracks[0].getFirstIndex()
        i = index.number
        # the first cut is the deepest
        counter = index.counter

        logger.debug('absolutizing')
        while True:
            track = self.tracks[t - 1]
            index = track.getIndex(i)
            assert track.number == t
            assert index.number == i
            if index.counter is None:
                logger.debug('track %d, index %d has no counter', t, i)
                break
            if index.counter != counter:
                logger.debug('track %d, index %d has a different counter',
                             t, i)
                break
            logger.debug('setting absolute offset %d on track %d, index %d',
                         index.relative, t, i)
            if index.absolute is not None:
                if index.absolute != index.relative:
                    msg = 'Track %d, index %d had absolute %d,' \
                        ' overriding with %d' % (
                            t, i, index.absolute, index.relative)
                    raise ValueError(msg)
            index.absolute = index.relative
            try:
                t, i = self.getNextTrackIndex(t, i)
            except IndexError:
                break

    def merge(self, other, session=2):
        """
        Merge the given table at the end.

        The other table is assumed to be from an additional session,

        :param other: session table
        :type other: Table
        :param session: session number
        :type session: int
        """
        gap = self._getSessionGap(session)

        trackCount = len(self.tracks)
        sourceCounter = self.tracks[-1].getLastIndex().counter

        for track in other.tracks:
            t = copy.deepcopy(track)
            t.number = track.number + trackCount
            t.session = session
            for i in list(t.indexes.values()):
                if i.absolute is not None:
                    i.absolute += self.leadout + gap
                    logger.debug('fixing track %02d, index %02d, absolute %d',
                                 t.number, i.number, i.absolute)
                if i.counter is not None:
                    i.counter += sourceCounter
                    logger.debug('fixing track %02d, index %02d, counter %d',
                                 t.number, i.number, i.counter)
            self.tracks.append(t)

        self.leadout += other.leadout + gap  # FIXME
        logger.debug('fixing leadout, now %d', self.leadout)

    @staticmethod
    def _getSessionGap(session):
        # From cdrecord multi-session info:
        # For the first additional session this is 11250 sectors
        # lead-out/lead-in overhead + 150 sectors for the pre-gap of the first
        # track after the lead-in = 11400 sectos.

        # For all further session this is 6750 sectors lead-out/lead-in
        # overhead + 150 sectors for the pre-gap of the first track after the
        # lead-in = 6900 sectors.

        gap = 11400
        if session > 2:
            gap = 6900
        return gap

    # lookups

    def getNextTrackIndex(self, track, index):
        """
        Return the next track and index.

        :param track: track number, 1-based
        :type track: int
        :raises IndexError: on last index
        :rtype: tuple(int, int)
        :param index: index of the next track
        :type index: int
        """
        t = self.tracks[track - 1]
        indexes = list(t.indexes)
        position = indexes.index(index)

        if position + 1 < len(indexes):
            return track, indexes[position + 1]

        track += 1
        if track > len(self.tracks):
            raise IndexError("No index beyond track %d, index %d" % (
                track - 1, index))

        t = self.tracks[track - 1]
        indexes = list(t.indexes)

        return track, indexes[0]

    # various tests for types of Table

    def hasTOC(self):
        """
        Check if the Table has a complete TOC.

        A TOC is a list of all tracks and their Index 01, with absolute
        offsets, as well as the leadout.
        """
        if not self.leadout:
            logger.debug('no leadout, no TOC')
            return False

        for t in self.tracks:
            if 1 not in list(t.indexes):
                logger.debug('no index 1, no TOC')
                return False
            if t.indexes[1].absolute is None:
                logger.debug('no absolute index 1, no TOC')
                return False

        return True

    def accuraterip_ids(self):
        """
        Return both AccurateRip disc ids.

        :returns: both AccurateRip disc ids as a tuple of 8-char
                  hexadecimal strings
        :rtype: tuple(str, str)
        """
        # AccurateRip does not take into account data tracks,
        # but does count the data track to determine the leadout offset
        discId1 = 0
        discId2 = 0

        for track in self.tracks:
            if not track.audio:
                continue
            offset = self.getTrackStart(track.number)
            discId1 += offset
            discId2 += (offset or 1) * track.number

        # also add end values, where leadout offset is one past the end
        # of the last track
        offset = self.getTrackEnd(self.tracks[-1].number) + 1
        discId1 += offset
        discId2 += offset * (self.getAudioTracks() + 1)

        discId1 &= 0xffffffff
        discId2 &= 0xffffffff

        return "%08x" % discId1, "%08x" % discId2

    def accuraterip_path(self):
        discId1, discId2 = self.accuraterip_ids()
        return "%s/%s/%s/dBAR-%.3d-%s-%s-%s.bin" % (
            discId1[-1], discId1[-2], discId1[-3],
            self.getAudioTracks(), discId1, discId2, self.getCDDBDiscId()
        )

    def canCue(self):
        """Check if this table can be used to generate a .cue file."""
        if not self.hasTOC():
            logger.debug('no TOC, cannot cue')
            return False

        for t in self.tracks:
            for i in list(t.indexes.values()):
                if i.relative is None:
                    logger.debug('track %02d, Index %02d does not '
                                 'have relative', t.number, i.number)
                    return False

        return True



# -*- Mode: Python; test-case-name: whipper.test.test_common_checksum -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

import binascii
import wave
import tempfile
import subprocess
import os


from whipper.extern.task import task as etask

import logging
logger = logging.getLogger(__name__)

# checksums are not CRC's. a CRC is a specific type of checksum.


class CRC32Task(etask.Task):
    # TODO: Support sampleStart, sampleLength later on (should be trivial, just
    # add change the read part in _crc32 to skip some samples and/or not
    # read too far)
    def __init__(self, path, sampleStart=0, sampleLength=-1, is_wave=True):
        self.path = path
        self.is_wave = is_wave

    def start(self, runner):
        etask.Task.start(self, runner)
        self.schedule(0.0, self._crc32)

    def _crc32(self):
        if not self.is_wave:
            _, tmpf = tempfile.mkstemp()

            try:
                subprocess.check_call(['flac', '-d', self.path, '-fo', tmpf])

                w = wave.open(tmpf)
            finally:
                os.remove(tmpf)
        else:
            w = wave.open(self.path)

        d = w._data_chunk.read()

        self.checksum = binascii.crc32(d) & 0xffffffff
        self.stop()



# -*- Mode: Python; test-case-name: whipper.test.test_common_accurip -*-
# vi:si:et:sw=4:sts=4:ts=4

import sys
from io import StringIO
from os.path import dirname, join
from unittest import TestCase

from whipper.common.accurip import (
    calculate_checksums, get_db_entry, print_report, verify_result,
    _split_responses, EntryNotFound
)
from whipper.result.result import RipResult, TrackResult


class TestAccurateRipResponse(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = 'c/1/2/dBAR-002-0000f21c-00027ef8-05021002.bin'
        with open(join(dirname(__file__), cls.path[6:]), 'rb') as f:
            cls.entry = _split_responses(f.read())
        cls.other_path = '4/8/2/dBAR-011-0010e284-009228a3-9809ff0b.bin'

    def test_raises_entrynotfound_for_no_entry(self):
        with self.assertRaises(EntryNotFound):
            get_db_entry('definitely_a_404')

    def test_AccurateRipResponse_parses_correctly(self):
        responses = get_db_entry(self.path)
        self.assertEqual(len(responses), 2)

        self.assertEqual(responses[0].num_tracks, 2)
        self.assertEqual(responses[0].discId1, '0000f21c')
        self.assertEqual(responses[0].discId2, '00027ef8')
        self.assertEqual(responses[0].cddbDiscId, '05021002')
        self.assertEqual(responses[0].confidences[0], 12)
        self.assertEqual(responses[0].confidences[1], 20)
        self.assertEqual(responses[0].checksums[0], '284fc705')
        self.assertEqual(responses[0].checksums[1], '9cc1f32e')

        self.assertEqual(responses[1].num_tracks, 2)
        self.assertEqual(responses[1].discId1, '0000f21c')
        self.assertEqual(responses[1].discId2, '00027ef8')
        self.assertEqual(responses[1].cddbDiscId, '05021002')
        self.assertEqual(responses[1].confidences[0], 9)
        self.assertEqual(responses[1].confidences[1], 9)
        self.assertEqual(responses[1].checksums[0], 'dc77f9ab')
        self.assertEqual(responses[1].checksums[1], 'dd97d2c3')

# XXX: test arc.py


class TestCalculateChecksums(TestCase):
    def test_returns_none_for_bad_files(self):
        self.assertEqual(
            calculate_checksums(['/does/not/exist']),
            {'v1': [None], 'v2': [None]}
        )

    # TODO: test success when file exists


class TestVerifyResult(TestCase):
    @classmethod
    def setUpClass(cls):
        path = 'c/1/2/dBAR-002-0000f21c-00027ef8-05021002.bin'
        with open(join(dirname(__file__), path[6:]), 'rb') as f:
            cls.responses = _split_responses(f.read())
        cls.checksums = {
            'v1': ['284fc705', '9cc1f32e'],
            'v2': ['dc77f9ab', 'dd97d2c3'],
        }

    def setUp(self):
        self.result = RipResult()
        for n in range(1, 2+1):
            track = TrackResult()
            track.number = n
            self.result.tracks.append(track)

    def test_empty_result_returns_false(self):
        self.assertEqual(
            verify_result(RipResult(), self.responses, self.checksums),
            False
        )

    def test_empty_responses_returns_false(self):
        self.assertEqual(
            verify_result(self.result, [], self.checksums),
            False
        )

    # XXX: would this happen?
    def test_empty_checksums_returns_false(self):
        self.assertEqual(
            verify_result(self.result, self.responses, {}),
            False
        )

    def test_wrong_checksums_returns_false(self):
        self.assertEqual(
            verify_result(self.result, self.responses, {
                'v1': ['deadbeef', '89abcdef'],
                'v2': ['76543210', '01234567']
            }),
            False
        )

    def test_incomplete_checksums(self):
        self.assertEqual(
            verify_result(self.result, self.responses, {
                'v1': ['284fc705', '9cc1f32e'],
                'v2': [None, 'dd97d2c3'],
            }),
            True
        )
        self.assertEqual(
            verify_result(self.result, self.responses, {
                'v1': ['284fc705', None],
                'v2': ['dc77f9ab', 'dd97d2c3'],
            }),
            True
        )
        self.assertEqual(
            verify_result(self.result, self.responses, {
                'v1': ['284fc705', None],
                'v2': [None, 'dd97d2c3'],
            }),
            True
        )

    def test_matches_only_v1_or_v2_responses(self):
        self.assertEqual(
            verify_result(
                self.result, [self.responses[0]], self.checksums
            ),
            True
        )
        self.assertEqual(
            verify_result(
                self.result, [self.responses[1]], self.checksums
            ),
            True
        )

    def test_passes_with_htoa(self):
        htoa = TrackResult()
        htoa.number = 0
        self.result.tracks.append(htoa)
        self.assertEqual(
            verify_result(self.result, self.responses, self.checksums),
            True
        )

    def test_stores_accuraterip_results_on_result(self):
        self.assertEqual(
            verify_result(self.result, self.responses, self.checksums),
            True
        )
        self.assertEqual(self.result.tracks[0].AR, {
            'v1': {
                'CRC': '284fc705',
                'DBCRC': '284fc705',
                'DBConfidence': 12,
            },
            'v2': {
                'CRC': 'dc77f9ab',
                'DBCRC': 'dc77f9ab',
                'DBConfidence': 7
            },
            'DBMaxConfidence': 12,
            'DBMaxConfidenceCRC': '284fc705',
        })
        self.assertEqual(self.result.tracks[1].AR, {
            'v1': {
                'CRC': '9cc1f32e',
                'DBCRC': '9cc1f32e',
                'DBConfidence': 20,
            },
            'v2': {
                'CRC': 'dd97d2c3',
                'DBCRC': 'dd97d2c3',
                'DBConfidence': 7,
            },
            'DBMaxConfidence': 20,
            'DBMaxConfidenceCRC': '9cc1f32e',
        })


class TestAccurateRipReport(TestCase):
    def setUp(self):
        sys.stdout = StringIO()
        self.result = RipResult()
        track = TrackResult()
        track.number = 1
        track.AR = {
            'v1': {
                'CRC': '284fc705',
                'DBCRC': '284fc705',
                'DBConfidence': 12,
            },
            'v2': {
                'CRC': 'dc77f9ab',
                'DBCRC': 'dc77f9ab',
                'DBConfidence': 4,
            },
            'DBMaxConfidence': 12,
            'DBMaxConfidenceCRC': '284fc705',
        }
        self.result.tracks.append(track)

    def tearDown(self):
        sys.stdout = sys.__stdout__

    def test_report_no_result(self):
        track = TrackResult()
        track.number = 1
        self.result.tracks[0] = track
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: unknown          (error)\n'
        )

    def test_track_not_found(self):
        self.result.tracks[0].AR['DBMaxConfidence'] = None
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: rip NOT accurate (not found)            '
            ' v1 [284fc705], v2 [dc77f9ab], DB [notfound]\n'
        )

    def test_htoa_not_tracked(self):
        self.result.tracks[0].number = 0
        self.result.tracks[0].AR['v1']['CRC'] = None
        self.result.tracks[0].AR['v2']['CRC'] = None
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  0: unknown          (not tracked)\n'
        )

    def test_report_v1_only(self):
        self.result.tracks[0].AR['v2']['DBCRC'] = None
        self.result.tracks[0].AR['v2']['DBConfidence'] = None
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: rip accurate     (max confidence     12)'
            ' v1 [284fc705], v2 [dc77f9ab], DB [284fc705]\n'
        )

    def test_report_v2_only(self):
        self.result.tracks[0].AR['v1']['DBCRC'] = None
        self.result.tracks[0].AR['v1']['DBConfidence'] = None
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: rip accurate     (confidence   4 of  12)'
            ' v1 [284fc705], v2 [dc77f9ab], DB [dc77f9ab]\n'
        )

    def test_report_v1_and_v2_max_confidence(self):
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: rip accurate     (max confidence     12)'
            ' v1 [284fc705], v2 [dc77f9ab], DB [284fc705, dc77f9ab]\n'
        )

    def test_report_v1_and_v2(self):
        self.result.tracks[0].AR['DBMaxConfidence'] = 66
        print_report(self.result)
        self.assertEqual(
            sys.stdout.getvalue(),
            'track  1: rip accurate     (confidence  12 of  66)'
            ' v1 [284fc705], v2 [dc77f9ab], DB [284fc705, dc77f9ab]\n'
        )



# -*- Mode: Python; test-case-name: whipper.test.test_common_common -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.


import os
import os.path
import math
import re
import subprocess
import unicodedata

from whipper.extern import asyncsub

import logging
logger = logging.getLogger(__name__)

FRAMES_PER_SECOND = 75

SAMPLES_PER_FRAME = 588  # a sample is 2 16-bit values, left and right channel
WORDS_PER_FRAME = SAMPLES_PER_FRAME * 2
BYTES_PER_FRAME = SAMPLES_PER_FRAME * 4


class EjectError(SystemError):
    """Possibly eject the drive in command.main."""

    def __init__(self, device, *args):
        """
        Init EjectError.

        :param args: a tuple used by ``BaseException.__str__``
        :param device: device path to eject
        """
        self.args = args
        self.device = device


def msfToFrames(msf):
    """
    Convert a string value in MM:SS:FF to frames.

    :param msf: the MM:SS:FF value to convert
    :type msf: str
    :returns: number of frames
    :rtype: int
    """
    if ':' not in msf:
        return int(msf)

    m, s, f = msf.split(':')

    return 60 * FRAMES_PER_SECOND * int(m) \
        + FRAMES_PER_SECOND * int(s) \
        + int(f)


def framesToMSF(frames, frameDelimiter=':'):
    f = frames % FRAMES_PER_SECOND
    frames -= f
    s = (frames / FRAMES_PER_SECOND) % 60
    frames -= s * FRAMES_PER_SECOND
    m = frames / FRAMES_PER_SECOND / 60

    return "%02d:%02d%s%02d" % (m, s, frameDelimiter, f)


def framesToHMSF(frames):
    # cdparanoia style
    f = frames % FRAMES_PER_SECOND
    frames -= f
    s = (frames / FRAMES_PER_SECOND) % 60
    frames -= s * FRAMES_PER_SECOND
    m = (frames / FRAMES_PER_SECOND / 60) % 60
    frames -= m * FRAMES_PER_SECOND * 60
    h = frames / FRAMES_PER_SECOND / 60 / 60

    return "%02d:%02d:%02d.%02d" % (h, m, s, f)


def formatTime(seconds, fractional=3):
    """
    Nicely format time in a human-readable format, like HH:MM:SS.mmm.

    If fractional is zero, no seconds will be shown.
    If it is greater than 0, we will show seconds and fractions of seconds.
    As a side consequence, there is no way to show seconds without fractions.

    :param seconds: the time in seconds to format
    :type seconds: int or float
    :param fractional: how many digits to show for the fractional part of
                       seconds
    :type fractional: int
    :returns: a nicely formatted time string
    :rtype: string
    """
    chunks = []

    if seconds < 0:
        chunks.append('-')
        seconds = -seconds

    hour = 60 * 60
    hours = seconds / hour
    seconds %= hour

    minute = 60
    minutes = seconds / minute
    seconds %= minute

    chunk = '%02d:%02d' % (hours, minutes)
    if fractional > 0:
        chunk += ':%0*.*f' % (fractional + 3, fractional, seconds)

    chunks.append(chunk)

    return " ".join(chunks)


class MissingDependencyException(Exception):
    dependency = None

    def __init__(self, *args):
        self.args = args
        self.dependency = args[0]


class EmptyError(Exception):
    pass


class MissingFrames(Exception):
    """Less frames decoded than expected."""

    pass


def truncate_filename(path):
    """Truncate filename to the max. len. allowed by the path's filesystem."""
    p, f = os.path.split(os.path.normpath(path))
    f, e = os.path.splitext(f)
    # Get the filename length limit in bytes
    fn_lim = os.pathconf(p.encode('utf-8'), 'PC_NAME_MAX')
    f_max = fn_lim - len(e.encode('utf-8'))
    f = unicodedata.normalize('NFC', f)
    f_trunc = f.encode()[:f_max].decode('utf-8', errors='ignore')
    return os.path.join(p, f_trunc + e)


def shrinkPath(path):
    """
    Shrink a full path to a shorter version.

    Used to handle ``ENAMETOOLONG``.
    """
    parts = list(os.path.split(path))
    length = len(parts[-1])
    target = 127
    if length <= target:
        target = pow(2, int(math.log(length, 2))) - 1

    name, ext = os.path.splitext(parts[-1])
    target -= len(ext) + 1

    # split on space, then reassemble
    words = name.split(' ')
    length = 0
    pieces = []
    for word in words:
        if length + 1 + len(word) <= target:
            pieces.append(word)
            length += 1 + len(word)
        else:
            break

    name = " ".join(pieces)
    # ext includes period
    parts[-1] = '%s%s' % (name, ext)
    path = os.path.join(*parts)
    return path


def getRealPath(refPath, filePath):
    """
    Translate a .cue or .toc's FILE argument to an existing path.

    Does Windows path translation.

    Will look for the given file name, but with .flac and .wav as extensions.

    :param refPath: path to the file from which the track is referenced;
                    for example, path to the .cue file in the same directory
    :type refPath: str
    :type filePath: str
    """
    assert isinstance(filePath, str), "%r is not str" % filePath

    if os.path.exists(filePath):
        return filePath

    candidatePaths = []

    # .cue FILE statements can have Windows-style path separators, so convert
    # them as one possible candidate
    # on the other hand, the file may indeed contain a backslash in the name
    # on linux
    # FIXME: I guess we might do all possible combinations of splitting or
    #        keeping the slash, but let's just assume it's either Windows
    #        or linux
    # See https://thomas.apestaart.org/morituri/trac/ticket/107
    parts = filePath.split('\\')
    if parts[0] == '':
        parts[0] = os.path.sep
    tpath = os.path.join(*parts)

    for path in [filePath, tpath]:
        if path == os.path.abspath(path):
            candidatePaths.append(path)
        else:
            # if the path is relative:
            # - check relatively to the cue file
            # - check only the filename part relative to the cue file
            candidatePaths.append(os.path.join(
                os.path.dirname(refPath), path))
            candidatePaths.append(os.path.join(
                os.path.dirname(refPath), os.path.basename(path)))

    # Now look for .wav and .flac files, as .flac files are often named .wav
    for candidate in candidatePaths:
        noext, _ = os.path.splitext(candidate)
        for ext in ['wav', 'flac']:
            cpath = '%s.%s' % (noext, ext)
            if os.path.exists(cpath):
                return cpath

    raise KeyError("Cannot find file for %r" % filePath)


def getRelativePath(targetPath, collectionPath):
    """
    Get a relative path from the directory of collectionPath to targetPath.

    Used to determine the path to use in .cue/.m3u files.
    """
    logger.debug('getRelativePath: target %r, collection %r',
                 targetPath, collectionPath)

    targetDir = os.path.dirname(targetPath)
    collectionDir = os.path.dirname(collectionPath)
    if targetDir == collectionDir:
        logger.debug('getRelativePath: target and collection in same dir')
        return os.path.basename(targetPath)
    rel = os.path.relpath(
        targetDir + os.path.sep,
        collectionDir + os.path.sep)
    logger.debug('getRelativePath: target and collection '
                 'in different dir, %r', rel)
    return os.path.join(rel, os.path.basename(targetPath))


def validate_template(template, kind):
    """Raise exception if disc/track template includes invalid variables."""
    if kind == 'disc':
        matches = re.findall(r'%[^ABCDIMNRSTXcdrxy]', template)
    elif kind == 'track':
        matches = re.findall(r'%[^ABCDIMNRSTXacdnrstxy]', template)
    if '%' in template and matches:
        raise ValueError(kind + ' template string contains invalid '
                         'variable(s): {}'.format(', '.join(matches)))


class VersionGetter:
    """
    Get the version of a program.

    It is extracted by looking for it in command output according to a RegEX.
    """

    def __init__(self, dependency, args, regexp, expander):
        """
        Init VersionGetter.

        :param dependency: name of the dependency providing the program
        :param args: the arguments to invoke to show the version
        :type args: list(str)
        :param regexp: the regular expression to get the version
        :param expander: the expansion string for the version using the
                         regexp group dict
        """
        self._dep = dependency
        self._args = args
        self._regexp = regexp
        self._expander = expander

    def get(self):
        version = "(Unknown)"

        try:
            with asyncsub.Popen(self._args,
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, close_fds=True) as p:
                p.wait()
                output = asyncsub.recv_some(p, e=0, stderr=1).decode()
            vre = self._regexp.search(output)
            if vre:
                version = self._expander % vre.groupdict()
        except OSError as e:
            import errno
            if e.errno == errno.ENOENT:
                raise MissingDependencyException(self._dep)
            raise

        return version



# -*- Mode: Python -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
import sys

logger = logging.getLogger(__name__)


class TaskException(Exception):
    """Wrap an exception that happened during task execution."""

    exception = None  # original exception

    def __init__(self, exception, message=None):
        self.exception = exception
        self.exceptionMessage = message
        self.args = (exception, message, )

# lifted from flumotion log module


def _getExceptionMessage(exception, frame=-1, filename=None):
    """
    Return a short message based on an exception, useful for debugging.

    Tries to find where the exception was triggered.
    """
    import traceback

    stack = traceback.extract_tb(sys.exc_info()[2])
    if filename:
        stack = [f for f in stack if f[0].find(filename) > -1]

    # badly raised exceptions can come without a stack
    if stack:
        (filename, line, func, text) = stack[frame]
    else:
        (filename, line, func, text) = ('no stack', 0, 'none', '')

    exc = exception.__class__.__name__
    msg = ""
    # a shortcut to extract a useful message out of most exceptions
    # for now
    if str(exception):
        msg = ": %s" % str(exception)
    return "exception %(exc)s at %(filename)s:%(line)s: %(func)s()%(msg)s" \
        % locals()


class LogStub:
    """Stub for a log interface."""

    @staticmethod
    def log(message, *args):
        logger.info(message, *args)

    @staticmethod
    def debug(message, *args):
        logger.debug(message, *args)

    @staticmethod
    def warning(message, *args):
        logger.warning(message, *args)


class Task(LogStub):
    """
    Wrap a task in an asynchronous interface.

    Can be listened to for starting, stopping, description changes
    and progress updates.

    I communicate an error by setting self.exception to an exception and
    stopping myself from running.
    The listener can then handle the Task.exception.

    :cvar description: what am I doing
    :cvar exception: set if an exception happened during the task
                     execution. Will be raised through ``run()`` at the end
    """

    logCategory = 'Task'

    description = 'I am doing something.'

    progress = 0.0
    increment = 0.01
    running = False
    runner = None
    exception = None
    exceptionMessage = None
    exceptionTraceback = None

    _listeners = None

    # subclass methods
    def start(self, runner):
        """
        Start the task.

        Subclasses should chain up to me at the beginning.

        Subclass implementations should raise exceptions immediately in
        case of failure (using set(AndRaise)Exception) first, or do it later
        using those methods.

        If start doesn't raise an exception, the task should run until
        complete, or ``setException()`` and ``stop()``.
        """
        self.debug('starting')
        self.setProgress(self.progress)
        self.running = True
        self.runner = runner
        self._notifyListeners('started')

    def stop(self):
        """
        Stop the task.

        Also resets the runner on the task.

        Subclasses should chain up to me at the end.
        It is important that they do so in all cases, even when
        they ran into an exception of their own.

        Listeners will get notified that the task is stopped,
        whether successfully or with an exception.
        """
        self.debug('stopping')
        self.running = False
        if not self.runner:
            print('ERROR: stopping task which is already stopped')
            import traceback
            traceback.print_stack()
        self.runner = None
        self.debug('reset runner to None')
        self._notifyListeners('stopped')

    # base class methods
    def setProgress(self, value):
        """
        Notify about progress changes bigger than the increment.

        Called by subclass implementations as the task progresses.
        """
        if (value - self.progress > self.increment or
                value >= 1.0 or value == 0.0):
            self.progress = value
            self._notifyListeners('progressed', value)
            self.debug('notifying progress: %r on %r',
                       value, self.description)

    def setDescription(self, description):
        if description != self.description:
            self._notifyListeners('described', description)
            self.description = description

    # FIXME: unify?
    def setExceptionAndTraceback(self, exception):
        """
        Call this to set a synthetically created exception.

        Not one that was actually raised and caught.
        """
        import traceback

        stack = traceback.extract_stack()[:-1]
        (filename, line, func, text) = stack[-1]
        exc = exception.__class__.__name__
        msg = ""
        # a shortcut to extract a useful message out of most exceptions
        # for now
        if str(exception):
            msg = ": %s" % str(exception)
        line = ("exception %(exc)s at %(filename)s:%(line)s: "
                "%(func)s()%(msg)s" % locals())

        self.exception = exception
        self.exceptionMessage = line
        self.exceptionTraceback = traceback.format_exc()
        self.debug('set exception, %r' % self.exceptionMessage)
    # FIXME: remove
    setAndRaiseException = setExceptionAndTraceback

    def setException(self, exception):
        """Call this to set a caught exception on the task."""
        import traceback

        self.exception = exception
        self.exceptionMessage = _getExceptionMessage(exception)
        self.exceptionTraceback = traceback.format_exc()
        self.debug('set exception, %r, %r' % (
            exception, self.exceptionMessage))

    def schedule(self, delta, callable_task, *args, **kwargs):
        if not self.runner:
            print("ERROR: scheduling on a task that's altready stopped")
            import traceback
            traceback.print_stack()
            return
        self.runner.schedule(self, delta, callable_task, *args, **kwargs)

    def addListener(self, listener):
        """
        Add a listener for task status changes.

        Listeners should implement started, stopped, and progressed.
        """
        self.debug('Adding listener %r', listener)
        if not self._listeners:
            self._listeners = []
        self._listeners.append(listener)

    def _notifyListeners(self, methodName, *args, **kwargs):
        if self._listeners:
            for listener in self._listeners:
                method = getattr(listener, methodName)
                try:
                    method(self, *args, **kwargs)
                # FIXME: catching too general exception (Exception)
                except Exception as e:
                    self.setException(e)


# FIXME: should this become a real interface, like in zope ?
class ITaskListener:
    """An interface for objects listening to tasks."""
    # listener callbacks

    def progressed(self, task, value):
        """
        Implement me to be informed about progress.

        :param task: a task
        :type task: Task
        :param value: progress, from 0.0 to 1.0
        :type value: float
        """

    def described(self, task, description):
        """
        Implement me to be informed about description changes.

        :param task: a task
        :type task: Task
        :param description: description
        :type description: str
        """

    def started(self, task):
        """Implement me to be informed about the task starting."""

    def stopped(self, task):
        """
        Implement me to be informed about the task stopping.

        If the task had an error, task.exception will be set.
        """


# this is a Dummy task that can be used to test if this works at all
class DummyTask(Task):
    def start(self, runner):
        Task.start(self, runner)
        self.schedule(1.0, self._wind)

    def _wind(self):
        self.setProgress(min(self.progress + 0.1, 1.0))

        if self.progress >= 1.0:
            self.stop()
            return

        self.schedule(1.0, self._wind)


class BaseMultiTask(Task, ITaskListener):
    """
    I perform multiple tasks.

    :cvar tasks: the tasks to run
    :vartype tasks: list(Task)
    """

    description = 'Doing various tasks'
    tasks = None

    def __init__(self):
        self.tasks = []
        self._task = 0

    def addTask(self, task):
        """
        Add a task.

        :type task: Task
        """
        if self.tasks is None:
            self.tasks = []
        self.tasks.append(task)

    def start(self, runner):
        """
        Start tasks.

        Tasks can still be added while running. For example,
        a first task can determine how many additional tasks to run.
        """
        Task.start(self, runner)

        # initialize task tracking
        if not self.tasks:
            self.warning('no tasks')
        self._generic = self.description

        self.next()

    def next(self):
        """Start the next task."""
        try:
            # start next task
            task = self.tasks[self._task]
            self._task += 1
            self.debug('BaseMultiTask.next(): starting task %d of %d: %r',
                       self._task, len(self.tasks), task)
            self.setDescription("%s (%d of %d) ..." % (
                task.description, self._task, len(self.tasks)))
            task.addListener(self)
            task.start(self.runner)
            self.debug('BaseMultiTask.next(): started task %d of %d: %r',
                       self._task, len(self.tasks), task)
        # FIXME: catching too general exception (Exception)
        except Exception as e:
            self.setException(e)
            self.debug('Got exception during next: %r', self.exceptionMessage)
            self.stop()
            return

    # ITaskListener methods
    def started(self, task):
        pass

    def progressed(self, task, value):
        pass

    def stopped(self, task):  # noqa: D401
        """
        Subclasses should chain up to me at the end of their implementation.

        They should fall through to chaining up if there is an exception.
        """
        self.debug('BaseMultiTask.stopped: task %r (%d of %d)',
                   task, self.tasks.index(task) + 1, len(self.tasks))
        if task.exception:
            self.warning('BaseMultiTask.stopped: exception %r',
                         task.exceptionMessage)
            self.exception = task.exception
            self.exceptionMessage = task.exceptionMessage
            self.stop()
            return

        if self._task == len(self.tasks):
            self.debug('BaseMultiTask.stopped: all tasks done')
            self.stop()
            return

        # pick another
        self.debug('BaseMultiTask.stopped: pick next task')
        self.schedule(0, self.next)


class MultiSeparateTask(BaseMultiTask):
    """
    Perform multiple tasks.

    Track progress of each individual task, going back to 0 for each task.
    """

    description = 'Doing various tasks separately'

    def start(self, runner):
        self.debug('MultiSeparateTask.start()')
        BaseMultiTask.start(self, runner)

    def next(self):
        self.debug('MultiSeparateTask.next()')
        # start next task
        self.progress = 0.0  # reset progress for each task
        BaseMultiTask.next(self)

    # ITaskListener methods
    def progressed(self, task, value):
        self.setProgress(value)

    def described(self, description):
        self.setDescription("%s (%d of %d) ..." % (
            description, self._task, len(self.tasks)))


class MultiCombinedTask(BaseMultiTask):
    """
    Perform multiple tasks.

    Track progress as a combined progress on all tasks on task granularity.
    """

    description = 'Doing various tasks combined'
    _stopped = 0

    # ITaskListener methods
    def progressed(self, task, value):
        self.setProgress(float(self._stopped + value) / len(self.tasks))

    def stopped(self, task):
        self._stopped += 1
        self.setProgress(float(self._stopped) / len(self.tasks))
        BaseMultiTask.stopped(self, task)


class TaskRunner(LogStub):
    """
    Base class for task runners.

    Task runners should be reusable.
    """

    logCategory = 'TaskRunner'

    def run(self, task):
        """
        Run the given task.

        :type task: Task
        """
        raise NotImplementedError

    # methods for tasks to call
    def schedule(self, delta, callable_task, *args, **kwargs):
        """
        Schedule a single future call.

        Subclasses should implement this.

        :param delta: time in the future to schedule call for, in seconds.
        :type delta: float
        :param callable_task: a task
        :type callable_task: Task
        """
        raise NotImplementedError


class SyncRunner(TaskRunner, ITaskListener):
    """Run the task synchronously in an asyncio event loop."""

    def __init__(self, verbose=True):
        self._verbose = verbose
        self._longest = 0  # longest string shown; for clearing

    def run(self, task, verbose=None, skip=False):
        self.debug('run task %r', task)
        self._task = task
        self._verboseRun = self._verbose
        if verbose is not None:
            self._verboseRun = verbose
        self._skip = skip

        self._loop = asyncio.new_event_loop()
        self._task.addListener(self)
        # only start the task after going into the mainloop,
        # otherwise the task might complete before we are in it
        self._loop.call_soon(self._startWrap, self._task)
        self.debug('run loop')
        self._loop.run_forever()

        self.debug('done running task %r', task)
        if task.exception:
            # catch the exception message
            # FIXME: this gave a traceback in the logging module
            self.debug('raising TaskException for %r, %r' % (
                task.exceptionMessage, task.exceptionTraceback))
            msg = task.exceptionMessage
            if task.exceptionTraceback:
                msg += "\n" + task.exceptionTraceback
            raise TaskException(task.exception, message=msg)

    def _startWrap(self, task):
        # wrap task start such that we can report any exceptions and
        # never hang
        try:
            self.debug('start task %r' % task)
            task.start(self)
        # FIXME: catching too general exception (Exception)
        except Exception as e:
            # getExceptionMessage uses global exception state that doesn't
            # hang around, so store the message
            task.setException(e)
            self.debug('exception during start: %r', task.exceptionMessage)
            self.stopped(task)

    def schedule(self, task, delta, callable_task, *args, **kwargs):
        def c():
            try:
                callable_task(*args, **kwargs)
                return False
            except Exception as e:
                self.debug('exception when calling scheduled callable %r',
                           callable_task)
                task.setException(e)
                self.stopped(task)
                raise

        self._loop.call_later(delta, c)

    # ITaskListener methods
    def progressed(self, task, value):
        if not self._verboseRun:
            return

        self._report()

        if value >= 1.0:
            if self._skip:
                self._output('%s %3d %%' % (
                    self._task.description, 100.0))
            else:
                # clear with whitespace
                print(("%s\r" % (' ' * self._longest, )), end='')

    def _output(self, what, newline=False, ret=True):
        print(what, end='')
        print((' ' * (self._longest - len(what))), end='')
        if ret:
            print('\r', end='')
        if newline:
            print('')
        sys.stdout.flush()
        if len(what) > self._longest:
            self._longest = len(what)

    def described(self, task, description):
        if self._verboseRun:
            self._report()

    def stopped(self, task):
        self.debug('stopped task %r', task)
        self.progressed(task, 1.0)
        self._loop.stop()

    def _report(self):
        self._output('%s %3d %%' % (
            self._task.description, self._task.progress * 100.0))


if __name__ == '__main__':
    task = DummyTask()
    runner = SyncRunner()
    runner.run(task)



# -*- Mode: Python; test-case-name: whipper.test.test_common_program -*-
# vi:si:et:sw=4:sts=4:ts=4

# Copyright (C) 2009, 2010, 2011 Thomas Vander Stichele

# This file is part of whipper.
#
# whipper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# whipper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with whipper.  If not, see <http://www.gnu.org/licenses/>.

"""Common functionality and class for all programs using whipper."""

import musicbrainzngs
import re
import os
import shutil
import time

from tempfile import NamedTemporaryFile
from packaging.version import Version

from whipper.common import accurip, checksum, common, mbngs, path
from whipper.program import cdrdao, cdparanoia
from whipper.result import result
from whipper.image import image
from whipper.extern import freedb
from whipper.extern.task import task

import logging
logger = logging.getLogger(__name__)


# FIXME: should Program have a runner ?


class Program:
    """
    I maintain program state and functionality.

    :vartype metadata: mbngs.DiscMetadata
    :cvar result: the rip's result
    :vartype result: result.RipResult
    :vartype outdir: str
    :vartype config: whipper.common.config.Config
    """

    cuePath = None
    logPath = None
    metadata = None
    outdir = None
    result = None
    skipped_tracks = None

    def __init__(self, config, record=False):
        """
        Init Program.

        :param record: whether to record results of API calls for playback
        """
        self._record = record
        self._config = config

        d = {}

        for key, default in list({
            'dot': True,
            'posix': True,
            'vfat': False,
            'whitespace': False,
            'printable': False
        }.items()):
            value = None
            value = self._config.getboolean('main', 'path_filter_' + key)
            if value is None:
                value = default

            d[key] = value

        self._filter = path.PathFilter(**d)

    @staticmethod
    def setWorkingDirectory(workingDirectory):
        if workingDirectory:
            logger.info('changing to working directory %s', workingDirectory)
            os.chdir(workingDirectory)

    def getFastToc(self, runner, device):
        """
        Retrieve the normal TOC table from the drive.

        Also warn about buggy cdrdao versions.
        """
        version = cdrdao.version()
        if Version(version) < Version('1.2.3rc2'):
            logger.warning('cdrdao older than 1.2.3 has a pre-gap length bug.'
                           ' See http://sourceforge.net/tracker/?func=detail&aid=604751&group_id=2171&atid=102171')  # noqa: E501

        t = cdrdao.ReadTOCTask(device, fast_toc=True)
        runner.run(t)
        toc = t.toc.table

        assert toc.hasTOC()
        return toc

    def getTable(self, runner, cddbdiscid, mbdiscid, device, offset,
                 toc_path):
        """
        Retrieve the Table from the drive.

        :rtype: table.Table
        """
        itable = None
        tdict = {}

        t = cdrdao.ReadTOCTask(device, toc_path=toc_path)
        t.description = "Reading table"
        runner.run(t)
        itable = t.toc.table
        tdict[offset] = itable
        logger.debug('getTable: read table %r', itable)

        assert itable.hasTOC()

        self.result.table = itable

        logger.debug('getTable: returning table with mb id %s',
                     itable.getMusicBrainzDiscId())
        return itable

    def getRipResult(self):
        """
        Return a new RipResult.

        :rtype: result.RipResult
        """
        assert self.result is None
        self.result = result.RipResult()
        return self.result

    @staticmethod
    def addDisambiguation(template_part, metadata):
        """Add disambiguation to template path part string."""
        if metadata.catalogNumber:
            template_part += ' (%s)' % metadata.catalogNumber
        elif metadata.barcode:
            template_part += ' (%s)' % metadata.barcode
        return template_part

    def getPath(self, outdir, template, mbdiscid, metadata, track_number=None):
        """
        Return disc or track path relative to outdir according to template.

        Track paths do not include extension.

        Tracks are named according to the track template, filling in
        the variables and adding the file extension. Variables
        exclusive to the track template are:

        * ``%t``: track number
        * ``%a``: track artist
        * ``%n``: track title
        * ``%s``: track sort name

        Disc files (.cue, .log, .m3u) are named according to the disc
        template, filling in the variables and adding the file
        extension. Variables for both disc and track template are:

        * ``%A``: release artist
        * ``%S``: release artist sort name
        * ``%B``: release barcode
        * ``%C``: release catalog number
        * ``%c``: release disambiguation comment
        * ``%d``: release title (with disambiguation)
        * ``%D``: disc title (without disambiguation)
        * ``%I``: MusicBrainz Disc ID
        * ``%M``: total number of discs in the chosen release
        * ``%N``: number of current disc
        * ``%T``: medium title
        * ``%y``: release year
        * ``%r``: release type, lowercase
        * ``%R``: release type, normal case
        * ``%x``: audio extension, lowercase
        * ``%X``: audio extension, uppercase
        """
        assert isinstance(outdir, str), "%r is not str" % outdir
        assert isinstance(template, str), "%r is not str" % template
        v = {}
        v['A'] = 'Unknown Artist'
        v['I'] = v['d'] = v['D'] = mbdiscid  # fallback for title
        v['r'] = 'unknown'
        v['R'] = 'Unknown'
        v['B'] = ''  # barcode
        v['C'] = ''  # catalog number
        v['x'] = 'flac'
        v['X'] = v['x'].upper()
        v['y'] = '0000'
        if track_number is not None:
            v['a'] = v['A']
            v['t'] = '%02d' % track_number
            if track_number == 0:
                v['n'] = 'Hidden Track One Audio'
            else:
                v['n'] = 'Unknown Track %d' % track_number

        if metadata:
            release = metadata.release or '0000'
            v['y'] = release[:4]
            v['A'] = metadata.artist
            v['S'] = metadata.sortName
            v['d'] = metadata.releaseTitle
            v['D'] = metadata.title
            v['B'] = metadata.barcode
            v['C'] = metadata.catalogNumber
            v['c'] = metadata.releaseDisambCmt
            v['M'] = metadata.discTotal
            v['N'] = metadata.discNumber
            v['T'] = metadata.mediumTitle
            if metadata.releaseType:
                v['R'] = metadata.releaseType
                v['r'] = metadata.releaseType.lower()
            if track_number is not None and track_number > 0:
                v['a'] = metadata.tracks[track_number - 1].artist
                v['s'] = metadata.tracks[track_number - 1].sortName
                v['n'] = metadata.tracks[track_number - 1].title
            elif track_number == 0:
                # htoa defaults to disc's artist
                v['a'] = metadata.artist

        template = re.sub(r'%(\w)', r'%(\1)s', template.strip('/'))
        # Avoid filtering non str type values, replace None with empty string
        v_fltr = {k: self._filter.filter(v2) if isinstance(v2, str) else ''
                  if v2 is None else v2 for k, v2 in v.items()}
        if outdir == os.curdir:
            return template % v_fltr  # Avoid useless './' in file paths
        return os.path.join(outdir, template % v_fltr)

    @staticmethod
    def getCDDB(cddbdiscid):
        """
        Fetch basic metadata from gnudb.org's mirror of freedb's CDDB.

        Freedb's official CDDB isn't used anymore because it's going to be
        shut down on 31/03/2020.

        See: https://web.archive.org/web/20200331093822/http://www.freedb.org/
        See: https://hydrogenaud.io/index.php?topic=118682

        :param cddbdiscid: list of id, tracks, offsets, seconds
        :rtype: str
        """
        # FIXME: convert to nonblocking?
        try:
            md = freedb.perform_lookup(
                     cddbdiscid, 'gnudb.gnudb.org', 80
            )
            logger.debug('CDDB query result: %r', md)
            return [item['DTITLE'] for item in md if 'DTITLE' in item] or None

        except ValueError as e:
            logger.warning("CDDB protocol error: %s", e)

        except IOError as e:
            # FIXME: for some reason errno is a str ?
            if e.errno == 'socket error':
                logger.warning("CDDB network error: %r", (e, ))
            else:
                raise

        return None

    def getMusicBrainz(self, ittoc, mbdiscid, release=None, country=None,
                       prompt=False):
        """
        Fetch MusicBrainz's metadata for the given MusicBrainz disc id.

        :param ittoc: disc TOC
        :type ittoc: whipper.image.table.Table
        :param mbdiscid: MusicBrainz DiscID
        :type mbdiscid: str
        :param release: MusicBrainz release id to match to
                        (if there are multiple)
        :type release: str or None
        :param country: country name used to filter releases by provenance
        :type country: str or None
        :param prompt: whether to prompt if there are multiple
                       matching releases
        :type prompt: bool
        """
        # look up disc on MusicBrainz
        print('Disc duration: %s, %d audio tracks' % (
            common.formatTime(ittoc.duration() / 1000.0),
            ittoc.getAudioTracks()))
        logger.debug('MusicBrainz submit url: %r',
                     ittoc.getMusicBrainzSubmitURL())

        metadatas = None

        for _ in range(0, 4):
            try:
                metadatas = mbngs.musicbrainz(mbdiscid,
                                              country=country,
                                              record=self._record)
                break
            except mbngs.NotFoundException as e:
                logger.warning("release not found: %r", (e, ))
                break
            except musicbrainzngs.NetworkError as e:
                logger.warning("network error: %r", (e, ))
                break
            except mbngs.MusicBrainzException as e:
                logger.warning("musicbrainz exception: %r", (e, ))
                time.sleep(5)
                continue

        if not metadatas:
            logger.warning('continuing without metadata')

        if metadatas:
            deltas = {}

            print('\nMatching releases:')

            for metadata in metadatas:
                print('\nArtist  : %s' % metadata.artist)
                print('Title   : %s' % metadata.releaseTitle)
                print('Duration: %s' % common.formatTime(
                                           metadata.duration / 1000.0))
                print('URL     : %s' % metadata.url)
                print('Release : %s' % metadata.mbid)
                print('Type    : %s' % metadata.releaseType)
                if metadata.barcode:
                    print("Barcode : %s" % metadata.barcode)
                if metadata.countries:
                    print("Country : %s" % ', '.join(metadata.countries))
                # TODO: Add test for non ASCII catalog numbers: see issue #215
                if metadata.catalogNumber:
                    print("Cat no  : %s" % metadata.catalogNumber)

                delta = abs(metadata.duration - ittoc.duration())
                if delta not in deltas:
                    deltas[delta] = []
                deltas[delta].append(metadata)

            lowest = None

            if not release and len(metadatas) > 1:
                # Select the release that most closely matches the duration.
                lowest = min(list(deltas))

                if prompt:
                    guess = (deltas[lowest])[0].mbid
                    release = input(
                        "\nPlease select a release [%s]: " % guess)

                    if not release:
                        release = guess

            if release:
                metadatas = [m for m in metadatas if m.url.endswith(release)]
                logger.debug('asked for release %r, only kept %r', release,
                             metadatas)
                if len(metadatas) == 1:
                    logger.info('picked requested release id %s', release)
                    print('Artist: %s' % metadatas[0].artist)
                    print('Title : %s' % metadatas[0].releaseTitle)
                elif not metadatas:
                    logger.warning("requested release id '%s', but none of "
                                   "the found releases match", release)
                    return None
            else:
                if lowest:
                    metadatas = deltas[lowest]

            # If we have multiple, make sure they match
            if len(metadatas) > 1:
                artist = metadatas[0].artist
                discTitle = metadatas[0].title
                for i, metadata in enumerate(metadatas):
                    if not artist == metadata.artist:
                        logger.warning("artist 0: %r and artist %d: %r are "
                                       "not the same", artist, i,
                                       metadata.artist)
                    if not discTitle == metadata.title:
                        logger.warning("title 0: %r and title %d: %r are "
                                       "not the same", discTitle, i,
                                       metadata.title)

                if not release and len(list(deltas)) > 1:
                    logger.warning('picked closest match in duration. '
                                   'Others may be wrong in MusicBrainz, '
                                   'please correct')
                    print('Artist : %s' % artist)
                    print('Title :  %s' % metadatas[0].title)

            # Select one of the returned releases. We just pick the first one.
            ret = metadatas[0]
        else:
            print('Submit this disc to MusicBrainz at the above URL.')
            ret = None

        print('')
        return ret

    def getTagList(self, number, mbdiscid):
        """
        Based on the metadata, get a dict of tags for the given track.

        :param number: track number (0 for HTOA)
        :type number: int
        :param mbdiscid: MusicBrainz DiscID
        :type mbdiscid: str
        :rtype: dict
        """
        trackArtist = 'Unknown Artist'
        releaseArtist = 'Unknown Artist'
        album = 'Unknown Album'
        title = 'Unknown Track'

        if self.metadata:
            trackArtist = self.metadata.artist
            releaseArtist = self.metadata.artist
            album = self.metadata.title  # No disambiguation is proper here
            mbidRelease = self.metadata.mbid
            mbidReleaseGroup = self.metadata.mbidReleaseGroup
            mbidReleaseArtist = self.metadata.mbidArtist

            if number > 0:
                try:
                    track = self.metadata.tracks[number - 1]
                    trackArtist = track.artist
                    title = track.title
                    mbidRecording = track.mbidRecording
                    mbidTrack = track.mbid
                    mbidTrackArtist = track.mbidArtist
                    mbidWorks = track.mbidWorks
                    composers = track.composers
                    performers = track.performers
                except IndexError as e:
                    logger.error('no track %d found, %r', number, e)
                    raise
            else:
                # htoa defaults to disc's artist
                title = 'Hidden Track One Audio'

        tags = {}

        if number > 0:
            tags['MUSICBRAINZ_DISCID'] = mbdiscid

        if self.metadata:
            tags['ALBUMARTIST'] = releaseArtist
        tags['ARTIST'] = trackArtist
        tags['TITLE'] = title
        tags['ALBUM'] = album

        tags['TRACKNUMBER'] = '%s' % number

        if self.metadata:
            if self.metadata.release is not None:
                tags['DATE'] = self.metadata.release
            if self.metadata.tracks:
                tags['TRACKTOTAL'] = str(len(self.metadata.tracks))
            if self.metadata.discTotal is not None:
                tags['DISCTOTAL'] = str(self.metadata.discTotal)
            if self.metadata.discNumber is not None:
                tags['DISCNUMBER'] = str(self.metadata.discNumber)

            if number > 0:
                tags['MUSICBRAINZ_RELEASETRACKID'] = mbidTrack
                tags['MUSICBRAINZ_TRACKID'] = mbidRecording
                tags['MUSICBRAINZ_ARTISTID'] = mbidTrackArtist
                tags['MUSICBRAINZ_ALBUMID'] = mbidRelease
                tags['MUSICBRAINZ_RELEASEGROUPID'] = mbidReleaseGroup
                tags['MUSICBRAINZ_ALBUMARTISTID'] = mbidReleaseArtist
                if len(mbidWorks) > 0:
                    tags['MUSICBRAINZ_WORKID'] = mbidWorks
                if len(composers) > 0:
                    tags['COMPOSER'] = composers
                if len(performers) > 0:
                    tags['PERFORMER'] = performers

        return tags

    def getHTOA(self):
        """
        Check if we have hidden track one audio.

        :returns: tuple of (start, stop), or None
        :rtype: tuple(int, int) or None
        """
        track = self.result.table.tracks[0]
        try:
            index = track.getIndex(0)
        except KeyError:
            return None

        start = index.absolute
        stop = track.getIndex(1).absolute - 1
        return start, stop

    def getCoverArt(self, path, release_id):
        """
        Get cover art image from Cover Art Archive.

        :param path: where to store the fetched image
        :type  path: str
        :param release_id: a release id (self.program.metadata.mbid)
        :type  release_id: str
        :returns: path to the downloaded cover art, else `None`
        :rtype: str or None
        """
        cover_art_path = os.path.join(path, 'cover.jpg')

        logger.debug('fetching cover art for release: %r', release_id)
        try:
            data = musicbrainzngs.get_image_front(release_id, 500)
        except musicbrainzngs.ResponseError as e:
            logger.error('error fetching cover art: %r', e)
            return

        if data:
            with NamedTemporaryFile(suffix='.cover.jpg', delete=False) as f:
                f.write(data)
            os.chmod(f.name, 0o644)
            shutil.move(f.name, cover_art_path)
            logger.debug('cover art fetched at: %r', cover_art_path)
            return cover_art_path
        return

    @staticmethod
    def verifyTrack(runner, trackResult):
        is_wave = not trackResult.filename.endswith('.flac')
        t = checksum.CRC32Task(trackResult.filename, is_wave=is_wave)

        try:
            runner.run(t)
        except task.TaskException as e:
            if isinstance(e.exception, common.MissingFrames):
                logger.warning('missing frames for %r', trackResult.filename)
                return False
            else:
                raise

        ret = trackResult.testcrc == t.checksum
        logger.debug('verifyTrack: track result crc %r, file crc %r, '
                     'result %r', trackResult.testcrc, t.checksum, ret)
        return ret

    def ripTrack(self, runner, trackResult, offset, device, taglist,
                 overread, what=None, coverArtPath=None):
        """
        Rip and store a track of the disc.

        Ripping the track may change the track's filename as stored in
        trackResult.

        :param runner: synchronous track rip task
        :type runner: task.SyncRunner
        :param trackResult: the object to store information in
        :type trackResult: result.TrackResult
        :param offset: ripping offset, in CD frames
        :type offset: int
        :param device: path to the hardware disc drive
        :type device: str
        :param taglist: dictionary of tags for the given track
        :type taglist: dict
        :param overread: whether to force overreading into the
                         lead-out portion of the disc
        :type overread: bool
        :param what: a string representing what's being read; e.g. Track
        :type what: str or None
        """
        if trackResult.number == 0:
            start, stop = self.getHTOA()
        else:
            start = self.result.table.getTrackStart(trackResult.number)
            stop = self.result.table.getTrackEnd(trackResult.number)

        dirname = os.path.dirname(trackResult.filename)
        os.makedirs(dirname, exist_ok=True)

        if not what:
            what = 'track %d' % (trackResult.number, )

        t = cdparanoia.ReadVerifyTrackTask(trackResult.filename,
                                           self.result.table, start,
                                           stop, overread,
                                           offset=offset,
                                           device=device,
                                           taglist=taglist,
                                           what=what,
                                           coverArtPath=coverArtPath)

        runner.run(t)

        logger.debug('ripped track')
        logger.debug('test speed %.3f/%.3f seconds',
                     t.testspeed, t.testduration)
        logger.debug('copy speed %.3f/%.3f seconds',
                     t.copyspeed, t.copyduration)
        trackResult.testcrc = t.testchecksum
        trackResult.copycrc = t.copychecksum
        trackResult.peak = t.peak
        trackResult.quality = t.quality
        trackResult.testspeed = t.testspeed
        trackResult.copyspeed = t.copyspeed
        # we want rerips to add cumulatively to the time
        trackResult.testduration += t.testduration
        trackResult.copyduration += t.copyduration

        if trackResult.filename != t.path:
            trackResult.filename = t.path
            logger.info('filename changed to %r', trackResult.filename)

    def verifyImage(self, runner, table):
        """
        Verify table against AccurateRip and cue_path track lengths.

        Verify our image against the given AccurateRip responses.

        Needs an initialized self.result.
        Will set accurip and friends on each TrackResult.

        Populates self.result.tracks with above TrackResults.
        """
        cueImage = image.Image(self.cuePath)
        # assigns track lengths
        if self.skipped_tracks is not None:
            verifytask = image.ImageVerifyTask(cueImage,
                                               [os.path.basename(t.filename)
                                                for t in self.skipped_tracks])
        else:
            verifytask = image.ImageVerifyTask(cueImage)
        runner.run(verifytask)
        if verifytask.exception:
            logger.error(verifytask.exceptionMessage)
            return False

        responses = accurip.get_db_entry(table.accuraterip_path())
        logger.info('%d AccurateRip response(s) found', len(responses))

        checksums = accurip.calculate_checksums([
            os.path.join(os.path.dirname(self.cuePath), t.indexes[1].path)
            for t in [t for t in cueImage.cue.table.tracks if t.number != 0]
        ])
        if not (checksums and any(checksums['v1']) and any(checksums['v2'])):
            return False

        return accurip.verify_result(self.result, responses, checksums)

    def write_m3u(self, discname):
        m3uPath = common.truncate_filename(discname + '.m3u')
        with open(m3uPath, 'w') as f:
            f.write('#EXTM3U\n')
            for track in self.result.tracks:
                if not track.filename:
                    # false positive htoa
                    continue
                if track.skipped:
                    continue
                if track.number == 0:
                    length = (self.result.table.getTrackStart(1) /
                              common.FRAMES_PER_SECOND)
                else:
                    length = (self.result.table.getTrackLength(track.number) /
                              common.FRAMES_PER_SECOND)

                target_path = common.getRelativePath(track.filename, m3uPath)
                u = '#EXTINF:%d,%s\n' % (length, target_path)
                f.write(u)
                u = '%s\n' % target_path
                f.write(u)

    def writeCue(self, discName):
        assert self.result.table.canCue()
        cuePath = common.truncate_filename(discName + '.cue')
        logger.debug('write .cue file to %s', cuePath)
        handle = open(cuePath, 'w')
        # FIXME: do we always want utf-8 ?
        handle.write(self.result.table.cue(cuePath))
        handle.close()

        self.cuePath = cuePath

        return cuePath

    def writeLog(self, discName, txt_logger):
        logPath = common.truncate_filename(discName + '.log')
        handle = open(logPath, 'w')
        log = txt_logger.log(self.result)
        handle.write(log)
        handle.close()

        self.logPath = logPath

        return logPath



# -*- Mode: Python; test-case-name: whipper.test.test_common_program -*-
# vi:si:et:sw=4:sts=4:ts=4


import os
import shutil
import unittest

from tempfile import NamedTemporaryFile
from whipper.common import program, mbngs, config
from whipper.command.cd import DEFAULT_DISC_TEMPLATE


class PathTestCase(unittest.TestCase):

    def testStandardTemplateEmpty(self):
        prog = program.Program(config.Config())

        path = prog.getPath('/tmp', DEFAULT_DISC_TEMPLATE,
                            'mbdiscid', None)
        self.assertEqual(path, ('/tmp/unknown/Unknown Artist - mbdiscid/'
                                'Unknown Artist - mbdiscid'))

    def testStandardTemplateFilled(self):
        prog = program.Program(config.Config())
        md = mbngs.DiscMetadata()
        md.artist = md.sortName = 'Jeff Buckley'
        md.releaseTitle = 'Grace'

        path = prog.getPath('/tmp', DEFAULT_DISC_TEMPLATE,
                            'mbdiscid', md, 0)
        self.assertEqual(path, ('/tmp/unknown/Jeff Buckley - Grace/'
                                'Jeff Buckley - Grace'))

    def testIssue66TemplateFilled(self):
        prog = program.Program(config.Config())
        md = mbngs.DiscMetadata()
        md.artist = md.sortName = 'Jeff Buckley'
        md.releaseTitle = 'Grace'

        path = prog.getPath('/tmp', '%A/%d', 'mbdiscid', md, 0)
        self.assertEqual(path,
                         '/tmp/Jeff Buckley/Grace')


# TODO: Test cover art embedding too.
class CoverArtTestCase(unittest.TestCase):

    @staticmethod
    def _mock_get_front_image(release_id):
        """
        Mock `musicbrainzngs.get_front_image` function.

        Reads a local cover art image and returns its binary data.

        :param release_id: a release id (self.program.metadata.mbid)
        :type  release_id: str
        :returns: the binary content of the local cover art image
        :rtype: bytes
        """
        filename = '%s.jpg' % release_id
        path = os.path.join(os.path.dirname(__file__), filename)
        with open(path, 'rb') as f:
            return f.read()

    def _mock_getCoverArt(self, path, release_id):
        """
        Mock `common.program.getCoverArt` function.

        :param path: where to store the fetched image
        :type  path: str
        :param release_id: a release id (self.program.metadata.mbid)
        :type  release_id: str
        :returns: path to the downloaded cover art
        :rtype: str
        """
        cover_art_path = os.path.join(path, 'cover.jpg')

        data = self._mock_get_front_image(release_id)

        with NamedTemporaryFile(suffix='.cover.jpg', delete=False) as f:
            f.write(data)
        os.chmod(f.name, 0o644)
        shutil.move(f.name, cover_art_path)
        return cover_art_path

    def testCoverArtPath(self):
        """Test whether a fetched cover art is saved properly."""
        # Using: Dummy by Portishead
        # https://musicbrainz.org/release/76df3287-6cda-33eb-8e9a-044b5e15ffdd
        path = os.path.dirname(__file__)
        release_id = "76df3287-6cda-33eb-8e9a-044b5e15ffdd"
        coverArtPath = self._mock_getCoverArt(path, release_id)
        self.assertTrue(os.path.isfile(coverArtPath))
